"""Workflow helpers used in the GEB."""

from time import time
from typing import Literal, overload

import numpy as np

from geb.geb_types import ArrayFloat
from geb.store import DynamicArray


class TimingModule:
    """A timing module to measure the time taken for different parts of a workflow."""

    def __init__(self, name: str) -> None:
        """Initializes the TimingModule with a name and starts the timer.

        Args:
            name: The name of the timing module. Will be used when printing the timing results.
        """
        self.name = name
        self.times = [time()]
        self.split_names = []

    def finish_split(self, name: str) -> None:
        """Finish split with with name given.

        Appends the current time and the name of the split to their respective lists, which will be
        used to calculate the time taken for each split and the total time when converting to string.

        Args:
            name: The name of the split. This is the name of the previous split.
        """
        self.times.append(time())
        self.split_names.append(name)

    def __str__(self) -> str:
        """Converts the timing information into a readable string format for logging or display.

        Returns:
            A formatted string summarizing the time taken for each split and the total time.
        """
        messages = []
        for i in range(1, len(self.times)):
            time_difference = self.times[i] - self.times[i - 1]
            messages.append(
                "{}: {:.4f}s".format(self.split_names[i - 1], time_difference)
            )

        # Calculate total time
        total_time = self.times[-1] - self.times[0]
        messages.append("Total: {:.4f}s".format(total_time))

        to_print = ", ".join(messages)
        to_print = "{} - {}".format(self.name, to_print)

        return to_print


@overload
def balance_check(
    name: str,
    how: str = "cellwise",
    influxes: list[ArrayFloat | np.floating | DynamicArray]
    | tuple[ArrayFloat | np.floating | DynamicArray] = [],
    outfluxes: list[ArrayFloat | np.floating | DynamicArray]
    | tuple[ArrayFloat | np.floating | DynamicArray] = [],
    prestorages: list[ArrayFloat | np.floating | DynamicArray]
    | tuple[ArrayFloat | np.floating | DynamicArray] = [],
    poststorages: list[ArrayFloat | np.floating | DynamicArray]
    | tuple[ArrayFloat | np.floating | DynamicArray] = [],
    tolerance: float = 1e-10,
    error_identifiers: dict = {},
    raise_on_error: bool = False,
    return_max_imbalance_index: Literal[False] = False,
) -> bool: ...


@overload
def balance_check(
    name: str,
    how: str = "cellwise",
    influxes: list[ArrayFloat | np.floating | DynamicArray]
    | tuple[ArrayFloat | np.floating | DynamicArray] = [],
    outfluxes: list[ArrayFloat | np.floating | DynamicArray]
    | tuple[ArrayFloat | np.floating | DynamicArray] = [],
    prestorages: list[ArrayFloat | np.floating | DynamicArray]
    | tuple[ArrayFloat | np.floating | DynamicArray] = [],
    poststorages: list[ArrayFloat | np.floating | DynamicArray]
    | tuple[ArrayFloat | np.floating | DynamicArray] = [],
    tolerance: float = 1e-10,
    error_identifiers: dict = {},
    raise_on_error: bool = False,
    return_max_imbalance_index: Literal[True] = ...,
) -> tuple[Literal[False], int] | tuple[Literal[True], None]: ...


def balance_check(
    name: str,
    how: str = "cellwise",
    influxes: list[ArrayFloat | np.floating | DynamicArray]
    | tuple[ArrayFloat | np.floating | DynamicArray] = [],
    outfluxes: list[ArrayFloat | np.floating | DynamicArray]
    | tuple[ArrayFloat | np.floating | DynamicArray] = [],
    prestorages: list[ArrayFloat | np.floating | DynamicArray]
    | tuple[ArrayFloat | np.floating | DynamicArray] = [],
    poststorages: list[ArrayFloat | np.floating | DynamicArray]
    | tuple[ArrayFloat | np.floating | DynamicArray] = [],
    tolerance: float = 1e-10,
    error_identifiers: dict = {},
    raise_on_error: bool = False,
    return_max_imbalance_index: bool = False,
) -> bool | tuple[Literal[False], int] | tuple[Literal[True], None]:
    """Check the balance of a system, usually for water.

    Essentially checks that influxes + prestorages = outfluxes + poststorages,
    within a given tolerance.

    Args:
        name: Name of the balance check, used for printing.
        how: Method to use for balance check, either 'cellwise' or 'sum'.
        influxes: List of influx arrays.
        outfluxes: List of outflux arrays.
        prestorages: List of pre-storage arrays.
        poststorages: List of post-storage arrays.
        tolerance: tolerance for the balance check.
        error_identifiers: Dictionary of identifiers to help locate errors, e.g. {'x': x_array, 'y': y_array}.
            Can only be used with how='cellwise'.
            When an error is found, the values of these identifiers at the location of the maximum error will be printed.
        raise_on_error: Whether to raise an error if the balance check fails.
        return_max_imbalance_index: Whether to return the maximum imbalance instead of a boolean.
            Only applicable when how='cellwise'.

    Returns:
        True if the balance check passes, False otherwise.

    Raises:
        ValueError: If NaN values are found in the balance calculation.
        AssertionError: If the balance check fails and raise_on_error is True.
        ValueError: If return_max_imbalance_index is True when using 'sum' method.
    """
    income = 0
    out = 0
    store = 0

    if how == "cellwise":
        influx = np.add.reduce(influxes)
        outflux = np.add.reduce(outfluxes)
        prestorage = np.add.reduce(prestorages)
        poststorage = np.add.reduce(poststorages)

        balance = influx - outflux + prestorage - poststorage

        if np.isnan(balance).any():
            for kind, array in zip(
                ["influx", "outflux", "prestorage", "poststorage"],
                [influxes, outfluxes, prestorages, poststorages],
            ):
                for i, component in enumerate(array, start=1):
                    if np.isnan(component).any():
                        raise ValueError(
                            f"NaN values found in {kind} component {i} (1-indexed)."
                        )
            else:
                raise ValueError(
                    "NaN values found in balance calculation, but could not identify component (shouldn't happen)."
                )

        if balance.size == 0:
            return True
        elif np.abs(balance).max() > tolerance:
            index = np.abs(balance).argmax()
            text = f"{balance[np.abs(balance).argmax()]} > tolerance {tolerance}, max imbalance at index {index}."

            if error_identifiers:
                text += " Error identifiers: " + ", ".join(
                    f"{key}={value[index]}" for key, value in error_identifiers.items()
                )
            if name:
                print(name, text)
            else:
                print(text)
            if raise_on_error:
                raise AssertionError(text)
            if return_max_imbalance_index:
                return False, index
            else:
                return False
        else:
            if return_max_imbalance_index:
                return True, None
            else:
                return True

    elif how == "sum":
        if return_max_imbalance_index:
            raise ValueError(
                "return_max_imbalance_index cannot be True when using 'sum' method."
            )
        assert not error_identifiers, (
            "Error identifiers not supported for 'sum' method."
        )
        for influx in influxes:
            income += influx.sum()
        for outflux in outfluxes:
            out += outflux.sum()
        for prestorage in prestorages:
            store += prestorage.sum()
        for poststorage in poststorages:
            store -= poststorage.sum()

        balance = abs(income + store - out)
        if np.isnan(balance):
            raise ValueError("Balance check failed, NaN values found.")
        if balance > tolerance:
            text = f"{balance} is larger than tolerance {tolerance}"
            if name:
                print(name, text)
            else:
                print(text)
            if raise_on_error:
                raise AssertionError(text)
            return False
        else:
            return True
    else:
        raise ValueError(f"Method {how} not recognized.")



"""Implementation of the Morgan–Morgan–Finney (MMF) hillslope erosion model."""

# --------------------------------------------------------------------------------
# This file contains code that has been adapted from an original source available
# in a public repository under the GNU General Public License. The original code
# has been modified to fit the specific needs of this project.
#
# Original source repository: https://github.com/FutureWater/SPHY
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------------------
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from geb.geb_types import ArrayFloat32
from geb.module import Module
from geb.store import Bucket

if TYPE_CHECKING:
    from geb.model import GEBModel, Hydrology


def get_kinetic_energy_direct_throughfall(
    direct_throughfall: npt.NDArray[np.float32],
    precipitation_intensity: npt.NDArray[np.float32],
) -> npt.NDArray[np.float32]:
    """Calculate the kinetic energy of rainfall.

    Args:
        direct_throughfall: Direct throughfall (mm).
        precipitation_intensity: Precipitation intensity derived from direct throughfall and alpha fraction (mm).

    Returns:
        Kinetic energy of rainfall (J/m²).
    """
    kinetic_energy = (
        direct_throughfall
        * (
            np.float32(0.29)
            * (
                np.float32(1.0)
                - np.float32(0.72) * np.exp(np.float32(-0.05) * precipitation_intensity)
            )
        )
        * np.float32(100.0)
    )  # Brown and Foster (1987)
    kinetic_energy[direct_throughfall == np.float32(0.0)] = np.float32(
        0.0
    )  # no kinetic energy if no direct_throughfall
    return kinetic_energy


def get_kinetic_energy_leaf_drainage(
    leaf_drainage: npt.NDArray[np.float32], plant_height: npt.NDArray[np.float32]
) -> npt.NDArray[np.float32]:
    """Calculate the kinetic energy of leaf drainage.

    Args:
        leaf_drainage: Leaf drainage (mm/day).
        plant_height: Plant height (m).

    Returns:
        Kinetic energy of leaf drainage (J/m²).
    """
    return np.where(
        plant_height >= np.float32(0.15),
        leaf_drainage
        * (np.float32(15.8) * plant_height ** np.float32(0.5) - np.float32(5.87)),
        np.float32(0.0),
    )


def get_detachment_from_raindrops(
    K: np.float32,
    texture_class_ratio: npt.NDArray[np.float32],
    kinetic_energy: npt.NDArray[np.float32],
    no_erosion: npt.NDArray[np.bool_],
    cover: npt.NDArray[np.float32],
) -> npt.NDArray[np.float32]:
    """Calculate soil detachment by raindrop impact.

    Args:
        K: Detachability of the soil by raindrop impact (g/J).
        texture_class_ratio: Ratio of texture class (clay, silt, or sand) (dimensionless).
        kinetic_energy: Kinetic energy of rainfall (J/m²).
        no_erosion: Mask indicating areas where erosion does not occur (dimensionless).
        cover: Ground cover fraction (dimensionless).

    Returns:
        Detachment from raindrops (kg/m²).
    """
    return (
        K
        * texture_class_ratio
        * np.maximum(
            np.float32(0.0), np.float32(1.0) - (no_erosion.astype(np.float32) + cover)
        )
        * kinetic_energy
        * np.float32(1e-3)
    )


def get_detachment_from_flow(
    DR: np.float32,
    texture_class_ratio: npt.NDArray[np.float32],
    direct_runoff: npt.NDArray[np.float32],
    slope: npt.NDArray[np.float32],
    no_erosion: npt.NDArray[np.bool_],
    cover: npt.NDArray[np.float32],
) -> npt.NDArray[np.float32]:
    """Calculate soil detachment by overland flow.

    Args:
        DR: Detachability of the soil by runoff (g/mm).
        texture_class_ratio: Ratio of texture class (clay, silt, or sand) (dimensionless).
        direct_runoff: Direct runoff (mm).
        slope: Slope angle (radians).
        no_erosion: Mask indicating areas where erosion does not occur (dimensionless).
        cover: Ground cover fraction (dimensionless).

    Returns:
        Detachment from flow (kg/m²).
    """
    return (
        DR
        * texture_class_ratio
        * direct_runoff ** np.float32(1.5)
        * np.maximum(
            np.float32(0.0), np.float32(1.0) - (no_erosion.astype(np.float32) + cover)
        )
        * np.sin(slope) ** np.float32(0.3)
        * np.float32(1e-3)
    )


def get_particle_fall_velocity(
    particle_diameter: np.float32,
    rho_s: np.float32,
    rho: np.float32,
    eta: np.float32,
    g: np.float32 = np.float32(9.81),
) -> np.float32:
    """Calculate particle fall velocity using Stokes' law.

    See https://doi.org/10.1002/esp.1530 equation 31.

    Args:
        particle_diameter: Particle diameter (m).
        rho_s: Sediment density (kg/m³).
        rho: Flow density (kg/m³).
        eta: Fluid viscosity (kg/(m·s)).
        g: Gravitational acceleration (m/s²), default is 9.81.

    Returns:
        Particle fall velocity (m/s).
    """
    return (
        np.float32(1.0 / 18.0)
        * (particle_diameter ** np.float32(2.0))
        * (rho_s - rho)
        * g
    ) / eta


def get_particle_fall_number(
    particle_diameter: np.float32,
    velocity: npt.NDArray[np.float32],
    water_depth: npt.NDArray[np.float32],
    rho_s: np.float32,
    rho: np.float32,
    eta: np.float32,
    slope: npt.NDArray[np.float32],
    cell_length: npt.NDArray[np.float32],
) -> npt.NDArray[np.float32]:
    """Calculate the particle fall number.

    See https://doi.org/10.1002/esp.1530 equations 28-30.

    Notes:
        It seems that the cell_length should not be used like this. Have a good look
        at this before using this in production.

    Args:
        particle_diameter: Particle diameter (m).
        velocity: Flow velocity (m/s).
        water_depth: Water depth (m).
        rho_s: Sediment density (kg/m³).
        rho: Flow density (kg/m³).
        eta: Fluid viscosity (kg/(m·s)).
        slope: Slope angle (radians).
        cell_length: Length of the cell (m).

    Returns:
        Particle fall number (dimensionless).
    """
    particle_fall_velocity = get_particle_fall_velocity(
        particle_diameter, rho_s, rho, eta
    )
    N_f = (cell_length / np.cos(slope) * particle_fall_velocity) / (
        velocity * water_depth
    )
    return N_f


def get_deposition(
    particle_fall_number: npt.NDArray[np.float32],
) -> npt.NDArray[np.float32]:
    """Calculate the percentage of particles deposited on the soil surface.

    Args:
        particle_fall_number: Particle fall number (dimensionless).

    Returns:
        Percentage of particles deposited on the soil surface (%).
    """
    return np.minimum(
        np.float32(44.1) * particle_fall_number ** np.float32(0.29), np.float32(100.0)
    )


def get_material_transport(
    detachment_from_raindrops: npt.NDArray[np.float32],
    detachment_from_flow: npt.NDArray[np.float32],
    deposition_percentage: npt.NDArray[np.float32],
) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]:
    """Calculate transported and redeposited material.

    Args:
        detachment_from_raindrops: Soil detached by raindrops (kg/m²).
        detachment_from_flow: Soil detached by flow (kg/m²).
        deposition_percentage: Percentage of material deposited (%).

    Returns:
        Tuple containing transported material (kg/m²) and redeposited material (kg/m²).
    """
    transported_material = (detachment_from_raindrops + detachment_from_flow) * (
        np.float32(1.0) - deposition_percentage / np.float32(100.0)
    )
    redeposited_material = (
        (detachment_from_raindrops + detachment_from_flow)
        * deposition_percentage
        / np.float32(100.0)
    )
    return transported_material, redeposited_material


def get_mannings_vegatation(
    water_depth: ArrayFloat32,
    stem_diameter: ArrayFloat32,
    number_of_stems: ArrayFloat32,
) -> ArrayFloat32:
    """Calculate the Manning's n for land with vegetation.

    Args:
        water_depth: Water depth (m).
        stem_diameter: Stem diameter (m).
        number_of_stems: Number of stems per m² (dimensionless).

    Returns:
        Manning's n for land with vegetation (dimensionless).
    """
    return (water_depth ** np.float32(0.67)) / (
        np.float32(2.0 * 9.81) / (stem_diameter * number_of_stems)
    ) ** np.float32(0.5)


def get_mannings_tillaged_soil(
    surface_roughness_parameter_tillage: np.float32,
) -> np.float32:
    """Calculate the Manning's n for tilled soil.

    Args:
        surface_roughness_parameter_tillage: Surface roughness parameter for tillage (dimensionless).

    Returns:
        Manning's n for tilled soil (dimensionless).
    """
    return np.exp(
        np.float32(-2.1132) + np.float32(0.0349) * surface_roughness_parameter_tillage
    )


def get_flow_velocity(
    manning: npt.NDArray[np.float32],
    water_depth: npt.NDArray[np.float32],
    slope: npt.NDArray[np.float32],
    minimum_slope: np.float32 = np.float32(1e-5),
) -> npt.NDArray[np.float32]:
    """Calculate the flow velocity using Manning's equation.

    Args:
        manning: Manning's coefficient (dimensionless).
        water_depth: Water depth (m).
        slope: Slope (m/m, dimensionless).
        minimum_slope: Minimum slope (m/m, dimensionless), default is 1e-5.

    Returns:
        Flow velocity (m/s).
    """
    return (
        np.float32(1.0)
        / manning
        * water_depth ** np.float32(2.0 / 3.0)
        * np.maximum(slope, minimum_slope) ** np.float32(0.5)
    )


class HillSlopeErosionVariables(Bucket):
    """Variables for the HillSlopeErosion module."""

    total_erosion: np.float32
    mannings_bare_soil: np.float32
    surface_roughness_parameter_tillage: np.float32
    alpha: np.float32
    K_clay: np.float32
    K_silt: np.float32
    K_sand: np.float32
    detachability_due_to_runoff_clay: np.float32
    detachability_due_to_runoff_silt: np.float32
    detachability_due_to_runoff_sand: np.float32
    particle_diameter_clay: np.float32
    particle_diameter_silt: np.float32
    particle_diameter_sand: np.float32
    rho: np.float32
    rho_s: np.float32
    eta: np.float32


class HillSlopeErosion(Module):
    """Calculate soil erosion on hillslopes using the MMF model.

    The Morgan–Morgan–Finney (MMF) model is a process-based soil erosion model
    developed by Morgan, Morgan, and Finney (1984). It is designed to estimate annual
    or event-based soil loss by considering the detachment and transport of soil
    particles separately.
    """

    var: HillSlopeErosionVariables

    def __init__(self, model: GEBModel, hydrology: Hydrology) -> None:
        """Initialize the hillslope erosion model.

        Currently in super alpha stage and uses a lot of constants and assumptions. Should NOT
        be used for science yet.

        Args:
            model: The GEB model instance.
            hydrology: The hydrology module instance.
        """
        super().__init__(model)
        self.hydrology = hydrology

        self.HRU = hydrology.HRU
        self.grid = hydrology.grid

        self.simulate = self.model.config["hazards"]["erosion"]["simulate"]

        if self.model.in_spinup:
            self.spinup()

    @property
    def name(self) -> str:
        """Return the name of the module.

        Returns:
            The name of the module.
        """
        return "hydrology.hillslope_erosion"

    def spinup(self) -> None:
        """Initialize variables needed for the hillslope erosion model.

        Currently this is in alpha stage and uses a lot of constants and assumptions. Should NOT
        be use for science yet.

        Raises:
            ValueError: For now always, until slope is checked.
        """
        if not self.simulate:
            return None

        self.var.total_erosion = np.float32(0.0)

        self.HRU.var.canopy_cover = self.HRU.full_compressed(
            np.float32(0.5), dtype=np.float32
        )  # using a constant for now
        self.HRU.var.ground_cover = self.HRU.full_compressed(
            np.float32(0.5), dtype=np.float32
        )  # using a constant for now
        self.HRU.var.plant_height = self.HRU.full_compressed(
            np.float32(1.0), dtype=np.float32
        )
        # need to see what this is, dependent on land use type probably.
        self.HRU.var.no_erosion = self.HRU.full_compressed(False, dtype=bool)
        # need to see what this is.
        self.HRU.var.cover = self.HRU.full_compressed(np.float32(0.0), dtype=np.float32)

        raise ValueError(
            "Check if slope is correctly implemented here. There seems to be some uncertainty whether slope should be in radians or m/m (or perhaps something else entirely)"
        )

        # Is correct? -> Does not seem to be correct. Should be "length" of the element. But what is that?
        raise ValueError(
            "Check if cell_length is correctly implemented here. Looking at the original code, cell length seems to implictly assume one hillslope element per cell. Which would have some implications for the dynamic cell size that we have in GEB"
        )
        self.HRU.var.cell_length = np.sqrt(self.HRU.var.cell_area)

        # NOTE: water depth in field seems quite deep now, and is not variable.
        # perhaps this should be made more dynamic?
        # Depth is dicussed here, depends on type of rill: https://doi.org/10.1002/esp.1530
        self.HRU.var.water_depth_in_field = self.HRU.full_compressed(
            np.float32(0.1), dtype=np.float32
        )
        self.HRU.var.stem_diameter = self.HRU.full_compressed(
            np.float32(0.01), dtype=np.float32
        )
        self.HRU.var.stem_diameter_harvested = self.HRU.full_compressed(
            np.float32(0.005), dtype=np.float32
        )

        self.var.mannings_bare_soil = np.float32(0.015)
        self.var.surface_roughness_parameter_tillage = np.float32(6.0)
        self.HRU.var.no_elements = self.HRU.full_compressed(
            np.float32(1000.0), dtype=np.float32
        )
        self.HRU.var.no_elements_harvested = self.HRU.full_compressed(
            np.float32(10000.0), dtype=np.float32
        )

        self.HRU.var.tillaged = self.HRU.full_compressed(True, dtype=bool)
        self.HRU.var.no_vegetation = self.HRU.full_compressed(False, dtype=bool)

        self.var.alpha = np.float32(0.34)  # using a constant for now

        # Detachability (K) of the soil by raindrop impact (g/J)
        self.var.K_clay = np.float32(0.1)
        self.var.K_silt = np.float32(0.5)
        self.var.K_sand = np.float32(0.3)

        # Detachability of the soil by runoff (g/mm)
        self.var.detachability_due_to_runoff_clay = np.float32(1.0)
        self.var.detachability_due_to_runoff_silt = np.float32(1.6)
        self.var.detachability_due_to_runoff_sand = np.float32(1.5)

        # particle diameter (m)
        self.var.particle_diameter_clay = np.float32(2e-6)
        self.var.particle_diameter_silt = np.float32(6e-5)
        self.var.particle_diameter_sand = np.float32(2e-4)

        self.var.rho = np.float32(
            1100.0
        )  # Flow density (kg/m³). Typically 1100 kg/m³ for runoff on hillslopes (Abrahams et al., 2001).
        self.var.rho_s = np.float32(
            2650.0
        )  # Sediment density (kg/m³). Typically 2650 kg/m³.
        self.var.eta = np.float32(
            0.0015
        )  # Fluid viscosity (kg/(m·s)). Nominally 0.001 kg/(m·s) but taken as 0.0015 to allow for the effects of the sediment in the flow.

    def get_flow_velocity(self) -> ArrayFloat32:
        """Calculate the flow velocity using Manning's equation.

        Returns:
            The flow velocity (m/s).
        """
        mannings_vegated_field = get_mannings_vegatation(
            self.HRU.var.water_depth_in_field,
            self.HRU.var.stem_diameter,
            self.HRU.var.no_elements,
        )
        mannings_vegated_field[self.HRU.var.no_vegetation] = np.float32(0.0)
        mannings_vegated_field[self.HRU.var.no_erosion] = np.float32(0.0)

        # this one is not yet implemented
        # mannings_vegated_field = pcr.ifthenelse(
        #     self.n_table > 0, self.n_table, mannings_vegated_field
        # )

        mannings_tilled_soil = get_mannings_tillaged_soil(
            self.var.surface_roughness_parameter_tillage
        )
        mannings_soil = np.where(
            self.HRU.var.tillaged, mannings_tilled_soil, self.var.mannings_bare_soil
        )

        mannings_field = (mannings_soil**2 + mannings_vegated_field**2) ** 0.5
        flow_velocity_field = get_flow_velocity(
            mannings_field,
            self.HRU.var.water_depth_in_field,
            self.HRU.var.slope_m_per_m,
        )

        mannings_vegated_field_harvested = get_mannings_vegatation(
            self.HRU.var.water_depth_in_field,
            self.HRU.var.stem_diameter_harvested,
            self.HRU.var.no_elements_harvested,
        )
        mannings_vegated_field_harvested[self.HRU.var.tillaged] = np.float32(0.0)

        mannings_field_harvested = (
            mannings_soil**2 + mannings_vegated_field_harvested**2
        ) ** 0.5
        flow_velocity_field_harvested = get_flow_velocity(
            mannings_field_harvested,
            self.HRU.var.water_depth_in_field,
            self.HRU.var.slope_m_per_m,
        )

        # all cells without a crop (-1), but with a land owner (not -1) can be considered harvested
        harvested = (self.HRU.var.crop_map == -1) & (self.HRU.var.land_owners != -1)
        flow_velocity = np.where(
            harvested,
            flow_velocity_field_harvested,
            flow_velocity_field,
        )
        return flow_velocity

    def step(self, overland_runoff_m: ArrayFloat32) -> ArrayFloat32 | None:
        """Perform one timestep of the hillslope erosion model.

        Returns:
            transported_material_kg: The amount of material transported in kg.
        """
        if not self.simulate:
            return None

        pr_mm_day = self.HRU.pr_kg_per_m2_per_s.mean() * np.float32(
            24.0 * 3600.0
        )  # kg/m²/s to mm/day (assuming water density = 1000 kg/m³)
        effective_rainfall_mm_day = pr_mm_day * np.cos(
            self.HRU.var.slope_m_per_m
        )  # slope-corrected rainfall (mm/day)

        leaf_drainage = effective_rainfall_mm_day * self.HRU.var.canopy_cover

        direct_throughfall = effective_rainfall_mm_day - leaf_drainage

        precipitation_intensity = (
            direct_throughfall * self.var.alpha
        )  # alpha is the fraction of rain in the highest intensity class (dimensionless)

        kinetic_energy_direct_throughfall = get_kinetic_energy_direct_throughfall(
            direct_throughfall, precipitation_intensity
        )

        kinetic_energy_leaf_drainage = get_kinetic_energy_leaf_drainage(
            leaf_drainage, self.HRU.var.plant_height
        )

        kinetic_energy = (
            kinetic_energy_direct_throughfall + kinetic_energy_leaf_drainage
        )

        detachment_from_raindrops_clay = get_detachment_from_raindrops(
            self.var.K_clay,
            self.HRU.var.clay_percentage[0]
            / np.float32(100.0),  # only consider top layer
            kinetic_energy,
            self.HRU.var.no_erosion,
            self.HRU.var.cover,
        )

        detachment_from_raindrops_silt = get_detachment_from_raindrops(
            self.var.K_silt,
            self.HRU.var.silt_percentage[0]
            / np.float32(100.0),  # only consider top layer
            kinetic_energy,
            self.HRU.var.no_erosion,
            self.HRU.var.cover,
        )

        detachment_from_raindrops_sand = get_detachment_from_raindrops(
            self.var.K_sand,
            self.HRU.var.sand_percentage[0]
            / np.float32(100.0),  # only consider top layer
            kinetic_energy,
            self.HRU.var.no_erosion,
            self.HRU.var.cover,
        )

        # detachment_from_raindrops = (
        #     detachment_from_raindrops_clay
        #     + detachment_from_raindrops_silt
        #     + detachment_from_raindrops_sand
        # )

        detachment_from_flow_clay = get_detachment_from_flow(
            self.var.detachability_due_to_runoff_clay,
            self.HRU.var.clay_percentage[0]
            / np.float32(100.0),  # only consider top layer
            overland_runoff_m * 1000.0,  # convert m to mm
            self.HRU.var.slope_m_per_m,
            self.HRU.var.no_erosion,
            self.HRU.var.cover,
        )

        detachment_from_flow_silt = get_detachment_from_flow(
            self.var.detachability_due_to_runoff_silt,
            self.HRU.var.silt_percentage[0]
            / np.float32(100.0),  # only consider top layer
            overland_runoff_m * 1000.0,
            self.HRU.var.slope_m_per_m,
            self.HRU.var.no_erosion,
            self.HRU.var.cover,
        )

        detachment_from_flow_sand = get_detachment_from_flow(
            self.var.detachability_due_to_runoff_sand,
            self.HRU.var.sand_percentage[0]
            / np.float32(100.0),  # only consider top layer
            overland_runoff_m * 1000.0,
            self.HRU.var.slope_m_per_m,
            self.HRU.var.no_erosion,
            self.HRU.var.cover,
        )

        # detachment_from_flow = (
        #     detachment_from_flow_clay
        #     + detachment_from_flow_silt
        #     + detachment_from_flow_sand
        # )

        velocity = self.get_flow_velocity()

        particle_fall_number_clay = get_particle_fall_number(
            self.var.particle_diameter_clay,
            velocity,
            self.HRU.var.water_depth_in_field,
            self.var.rho_s,
            self.var.rho,
            self.var.eta,
            self.HRU.var.slope_m_per_m,
            self.HRU.var.cell_length,
        )

        particle_fall_number_silt = get_particle_fall_number(
            self.var.particle_diameter_silt,
            velocity,
            self.HRU.var.water_depth_in_field,
            self.var.rho_s,
            self.var.rho,
            self.var.eta,
            self.HRU.var.slope_m_per_m,
            self.HRU.var.cell_length,
        )

        particle_fall_number_sand = get_particle_fall_number(
            self.var.particle_diameter_sand,
            velocity,
            self.HRU.var.water_depth_in_field,
            self.var.rho_s,
            self.var.rho,
            self.var.eta,
            self.HRU.var.slope_m_per_m,
            self.HRU.var.cell_length,
        )

        percentage_deposition_clay = get_deposition(particle_fall_number_clay)
        percentage_deposition_silt = get_deposition(particle_fall_number_silt)
        percentage_deposition_sand = get_deposition(particle_fall_number_sand)

        transported_material_clay, redeposited_material_clay = get_material_transport(
            detachment_from_raindrops_clay,
            detachment_from_flow_clay,
            percentage_deposition_clay,
        )
        transported_material_silt, redeposited_material_silt = get_material_transport(
            detachment_from_raindrops_silt,
            detachment_from_flow_silt,
            percentage_deposition_silt,
        )
        transported_material_sand, redeposited_material_sand = get_material_transport(
            detachment_from_raindrops_sand,
            detachment_from_flow_sand,
            percentage_deposition_sand,
        )

        transported_material = (
            transported_material_clay
            + transported_material_silt
            + transported_material_sand
        )
        # redeposited_material = redeposited_material_clay + redeposited_material_silt + redeposited_material_sand  # Is D the total material deposited?

        transported_material_kg = transported_material * self.HRU.var.cell_area  # kg

        self.var.total_erosion += transported_material_kg.sum()

        self.report(locals())

        return transported_material



"""Contains classes and methods for building the dependency tree of build methods, verification etc."""

import functools
import inspect
import logging
from pathlib import Path
from time import time
from typing import Any, Iterable

import matplotlib.pyplot as plt
import networkx as nx

__all__: list[str] = ["build_method"]

from typing import Protocol


class NamedCallable(Protocol):
    __name__: str

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


class _build_method:
    def __init__(self, logger: logging.Logger | None = None) -> None:
        self.logger: logging.Logger | None = logger
        self.tree = nx.DiGraph()
        self.required_methods: set[str] = set()
        self.time_taken: dict[str, float] = {}

    def _resolve_logger(
        self, call_args: tuple[Any, ...] | None = None
    ) -> logging.Logger:
        """Resolve the logger for build method logging.

        Uses the logger on the bound model instance when available.
        For module-level operations, the logger must be assigned explicitly
        (for example via `build_method.logger = instance.logger`).

        Args:
            call_args: Positional arguments passed to a wrapped build method.

        Returns:
            Logger to use for logging build method messages.

        Raises:
            RuntimeError: If no logger is available.
        """
        if call_args:
            instance: Any = call_args[0]
            instance_logger: Any = getattr(instance, "logger", None)
            if isinstance(instance_logger, logging.Logger):
                return instance_logger
        if self.logger is None:
            raise RuntimeError(
                "build_method logger is not set. Ensure the owning class sets "
                "build_method.logger before invoking build methods."
            )
        return self.logger

    def __call__(
        self,
        required: bool,
        func: NamedCallable | None = None,
        depends_on: str | list[str] | None = None,
    ) -> NamedCallable:
        """Decorator to mark a method as a build method.

        Args:
            required: Whether the method is required to run.
            func: The function to decorate.
            depends_on: A method name or list of build_method that this method depends on.

        Returns:
            The decorated function.

        Raises:
            TypeError: if the decorator is used without parentheses.
        """

        def partial_decorator(func: NamedCallable) -> NamedCallable:
            @functools.wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                active_logger: logging.Logger = self._resolve_logger(args)
                active_logger.info(f"Running method: {func.__name__}")
                for key, value in kwargs.items():
                    active_logger.debug(f"{func.__name__}.{key}: {value}")

                start_time: float = time()
                value: Any = func(*args, **kwargs)
                end_time: float = time()

                elapsed_time: float = end_time - start_time

                self.time_taken[func.__name__] = elapsed_time
                active_logger.info(
                    f"Completed {func.__name__} in {elapsed_time:.2f} seconds"
                )
                return value

            self.add_tree_node(func)

            if depends_on is not None:
                if isinstance(depends_on, str):
                    self.add_tree_edge(func, depends_on)
                elif isinstance(depends_on, list):
                    for item in depends_on:
                        self.add_tree_edge(func, item)
                else:
                    raise ValueError("depends_on must be a string or a list of strings")

            if required:
                self.required_methods.add(func.__name__)

            setattr(wrapper, "__is_build_method__", True)
            return wrapper

        if func is None:
            return partial_decorator
        else:
            raise TypeError(
                "Use @build_method() rather than @build_method without parentheses."
            )

    def add_tree_node(self, func: NamedCallable) -> None:
        """Add a node to the dependency tree."""
        parameters = inspect.signature(func).parameters
        required_parameters = [
            param.name
            for param in parameters.values()
            if param.default is inspect.Parameter.empty and param.name != "self"
        ]
        optional_parameters = [
            param.name
            for param in parameters.values()
            if param.default is not inspect.Parameter.empty
        ]
        self.tree.add_node(
            func.__name__,
            attr={
                "_function_exists": True,
                "_required_parameters": required_parameters,
                "_optional_parameters": optional_parameters,
            },
        )

    def add_tree_edge(self, func: NamedCallable, depends_on: str) -> None:
        """Add an edge to the dependency tree.

        Raises:
            ValueError: if a method depends on "setup_region" since everything depends on it.
                "setup_region" should therefore not be included in the dependency tree.

        """
        if depends_on == "setup_region":
            raise ValueError(
                "Everything depends on setup_region so we don't include it."
            )
        self.tree.add_edge(depends_on, func.__name__)

    def validate_tree(self) -> None:
        """Validate the dependency tree.

        Checks if all the node dependencies are present in the tree.

        Raises:
            ValueError: if a method depends on another method that is not a build function.
        """
        assert nx.is_directed_acyclic_graph(self.tree)
        for method in self.tree.nodes:
            depencencies = list(self.tree.predecessors(method))
            for dependency in depencencies:
                if (
                    not self.tree.nodes[dependency]
                    .get("attr", {})
                    .get("_function_exists", False)
                ):
                    raise ValueError(
                        f"Method {method} depends on {dependency}, "
                        "which is not a build function."
                    )
        self._resolve_logger().debug("Builder dependency tree validation passed.")

    def validate_methods(
        self, methods: dict[str, Any], validate_order: bool = True
    ) -> None:
        """Validate the methods in the dependency tree.

        Currently check the order of methods and whether they have the correct parameters.

        Args:
            methods: A dictionary of methods to validate.
            validate_order: If True, checks if methods depend on other methods that may come after them in the build file.

        Raises:
            ValueError: If a method depends on another method that is not a build function
            ValueError: If a method depends on another method that may come after it in the build file or not be present at all.
            ValueError: If a method has parameters that do not match the expected parameters.
        """
        for method in methods:
            if not self.tree.has_node(method):
                raise ValueError(
                    f"Method {method} is not a build function.  If you are sure this should be build method, please decorate it with @build_method.'"
                )

            node = self.tree.nodes[method]
            if not node.get("attr", {}).get("_function_exists", False):
                raise ValueError(f"Method {method} is not a build function.")

        if validate_order:
            processed_methods = set()
            for method in methods:
                direct_dependencies = self.get_dependencies(
                    method, depth_limit=2
                )  # 1 is the method itself
                for direct_dependency in direct_dependencies:
                    if direct_dependency == method:
                        continue
                    if direct_dependency not in processed_methods:
                        if direct_dependency not in self.methods:
                            raise ValueError(
                                f"Method {method} depends on {direct_dependency}, "
                                "which is not a build function."
                            )
                        else:
                            raise ValueError(
                                f"Method {method} depends on {direct_dependency}, "
                                "which may come after this method in the build file or not be present at all."
                            )
                processed_methods.add(method)

        for method, args in methods.items():
            args_set = set(args) if args is not None else set()

            required_parameters = self.tree.nodes[method]["attr"][
                "_required_parameters"
            ]
            if not set(required_parameters).issubset(args_set):
                raise ValueError(
                    f"Method {method} has parameters {list(args.keys())}, "
                    f"but expected parameters are {required_parameters}."
                )

            optional_parameters = self.tree.nodes[method]["attr"][
                "_optional_parameters"
            ]
            if not set(args_set).issubset(
                set(required_parameters + optional_parameters)
            ):
                raise ValueError(
                    f"Method {method} has parameters {list(args.keys())}, "
                    f"but required parameters are {required_parameters} and "
                    f"optional parameters are {optional_parameters}."
                )

    def export_tree(self) -> None:
        pos = nx.spring_layout(self.tree)
        nx.draw(self.tree, pos, with_labels=True, arrows=True)
        plt.savefig("dependency_graph.png")

    def get_dependencies(
        self, method: str, depth_limit: None | int = None
    ) -> list[str]:
        """Get all dependencies for a given method.

        Args:
            method: The method for which to find dependencies.
            depth_limit: Optional depth limit for the search.

        Returns:
            A list of methods that the specified method depends on.
        """
        # [:-1] is used to skip the method itself in the result
        return list(
            nx.dfs_postorder_nodes(self.tree.reverse(), method, depth_limit=depth_limit)
        )[:-1]

    def get_dependents(self, method: str, depth_limit: None | int = None) -> list[str]:
        """Get all methods that depend on a given method.

        Args:
            method: The method for which to find dependents.
            depth_limit: Optional depth limit for the search.

        Returns:
            A list of methods that depend on the specified method.
        """
        # [1:] is used to skip the method itself in the result
        return list(nx.dfs_preorder_nodes(self.tree, method, depth_limit=depth_limit))[
            1:
        ]

    def record_progress(self, progress_path: Path, method: str) -> None:
        """Record progress to txt progress file.

        Args:
            progress_path: Path to the progress file.
            method: Method that has been completed.
        """
        with open(progress_path, "a") as f:
            f.write(f"{method}\n")

    def read_progress(self, progress_path: Path) -> list[str]:
        """Get the list of methods that have been completed from the progress file.

        Args:
            progress_path: Path to the progress file.

        Returns:
            A list of methods that have been completed.

        Raises:
            ValueError: If duplicate methods are found in the progress file.
        """
        if not progress_path.exists():
            return []
        with open(progress_path, "r") as f:
            completed_methods: list[str] = f.read().splitlines()

        # Check for duplicates
        seen: set[str] = set()
        duplicates: list[str] = []
        for method in completed_methods:
            if method in seen:
                duplicates.append(method)
            seen.add(method)

        if duplicates:
            raise ValueError(
                f"Progress file corrupted with duplicates: {duplicates}. Possibly you run two concurrent builds at the same time? "
                f"Remove duplicates from the progress.txt and restart build with --continue."
            )

        return completed_methods

    def check_required_methods(self, methods: Iterable[str]) -> None:
        """Check that all required methods are present in the provided method list.

        Args:
            methods: A list of method names to check.

        Raises:
            ValueError: If any required method is missing.
        """
        missing_methods: set[str] = self.required_methods - set(methods)
        if missing_methods:
            raise ValueError(
                f"The following required methods are missing: {', '.join(missing_methods)}"
            )

    def log_time_taken(self) -> None:
        """Log the time taken for each method in the dependency tree."""
        active_logger: logging.Logger = self._resolve_logger()
        total_time: float = sum(self.time_taken.values())

        sorted_by_time = sorted(
            self.time_taken.items(), key=lambda item: item[1], reverse=False
        )

        for method, time_taken in sorted_by_time:
            percentage: float = (time_taken / total_time) * 100
            active_logger.info(
                f"Method {method} took {time_taken:.2f} seconds ({percentage:.1f}%)."
            )

        active_logger.info(f"Total time taken: {total_time:.2f} seconds.")

    @property
    def methods(self) -> list[str]:
        """Return the methods in the dependency tree, sorted alphabetically.

        Returns:
            A alphabetically sorted list of method names in the dependency tree.
        """
        return sorted(list(self.tree.nodes))


build_method = _build_method()



"""This module is based on the HGRAPHER and COAST-RP modeling frameworks.

Source code can be found on https://doi.org/10.5281/zenodo.7912730.
"""

import itertools
import os
from datetime import datetime, timedelta
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.signal as ss
import scipy.stats as sp
from pandas.plotting import (
    register_matplotlib_converters,
)

from geb.workflows.io import read_geom

from ....workflows.io import read_table


def linear_detrend_with_nan(
    data: pd.Series,
    ref_date: datetime,
    station: int,
    remove_means: bool = True,
    figure_plotting: bool = True,
) -> tuple[float, pd.Series]:
    """This function detrends a time series linearly, while keeping NaN values in the original time series.

    The function can be used to remove long-term trends from the tide signal,
    which could affect the analysis of the tidal cycles and the surge hydrograph.

    Args:
        data: A pd.Series with date as index and sea level as values.
        ref_date: If a date is mentioned, remove trend taking the sealevel on this date as reference.
        station: The station identifier.
        remove_means: If True, centers the detrended timeseries around 0.
        figure_plotting: If True, returns a figure of both timeseries.
    Returns:
        The linearly detrended data with time as index.
    """
    y = np.array(data.values)  # sealevel time series
    x = np.arange(
        0, len(y), 1
    )  # create x-values ranging from 0 to 1 corresponding to the sealevel timeseries (y)
    not_nan_ind = ~np.isnan(
        y
    )  # create boolean, all indexes that are not nan become 'True'
    # regression line: m=slope; b=intercept; r_val=correlation coefficient; p_val=two-sided p-value; std_err=standard error of the estimated gradient
    m, b, r_val, p_val, std_err = sp.linregress(
        x[not_nan_ind], y[not_nan_ind]
    )  # see docs 'https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.linregress.html'

    if remove_means:  # if true, values will be centered around the mean, such that the mean of the complete time series becomes zero
        detrend_y = (
            y - (m * x + b)
        )  # y = original time series; m = slope regression line; x = x-values range; b = intercept of the regression line;
    elif (
        ref_date is not None
    ):  # elif ref_date is given, detrend with this date as reference point
        x_0 = np.flatnonzero(
            data.index == ref_date
        )  # index of the refdate used to detrend the complete time series
        detrend_y = (
            y - (m * x + b) + (m * x_0 + b)
        )  # y = original time series; m = slope regression line; x = x-values range; b = intercept of the regression line;
    else:  # else ref_date automatically becomes the first timestep of the time series
        detrend_y = y - (
            m * x
        )  # y = original time series; m = slope regression line; x = x-values range;

    if figure_plotting:
        register_matplotlib_converters()
        plt.figure()
        plt.plot(data.index, y, label="original")
        plt.plot(data.index, detrend_y, label="detrended")
        plt.xlabel("year")
        plt.ylabel("waterlevel [m]")
        plt.legend(loc="upper right")
        plt.savefig(
            "plot/gtsm/tide_detrended_station_%05d.png" % (int(station)),
            format="png",
            bbox_inches="tight",
            dpi=300,
        )
        plt.close()
    result = pd.Series(
        data=detrend_y,
        index=data.index,
    )  # create a pandas series with the detrended time series and dates as index
    return m, result


def generate_storm_surge_hydrographs(model: Any, make_plot: bool = False) -> None:
    """Generate storm surge hydrographs for a given GEB model.

    Args:
        model: The GEB model instance.
        make_plot: Whether to create plots of the hydrographs.
    """
    # read geojson file to get station ids
    station_ids = read_geom(model.files["geom"]["gtsm/stations_coast_rp"])
    os.makedirs("plot/gtsm", exist_ok=True)
    percentile = 0.99
    offset = 0
    rps = read_table(model.files["table"]["coast_rp"])
    waterlevels = read_table(model.files["table"]["gtsm/waterlevels"])
    surge = read_table(model.files["table"]["gtsm/surge"])
    return_periods = [int(rp) for rp in rps.columns.tolist()]
    df_event = {}
    df_event_spring = {}

    for station in station_ids["station_id"]:
        df_event[station] = {}
        df_event_spring[station] = {}
        waterlevelpd = waterlevels[int(station)]
        surgepd = surge[int(station)]
        tidepd = waterlevelpd - surgepd
        # detrend the tide signal to remove any long-term trends that might be present in the data, which could affect the analysis of the tidal cycles and the surge hydrograph
        _m, result = linear_detrend_with_nan(
            tidepd,
            tidepd.index[-1],
            station,
            remove_means=False,
            figure_plotting=make_plot,
        )
        tidepd = result
        average_tide_signal, spring_tide_signal = generate_tide_signals(
            station, tidepd, make_plot=make_plot
        )
        surge_hydrograph_height, surge_hydrograph_duration_mean = (
            generate_surge_hydrograph(station, surgepd, percentile, make_plot)
        )
        for rp in return_periods:
            df_event[station][rp], df_event_spring[station][rp] = (
                generate_storm_tide_hydrograph(
                    rps,
                    station,
                    tidepd,
                    surgepd,
                    average_tide_signal,
                    spring_tide_signal,
                    surge_hydrograph_duration_mean,
                    surge_hydrograph_height,
                    percentile,
                    rp,
                    offset,
                    make_plot,
                )
            )

    # save storm tide hydrograph (normal tide)
    export_folder = model.output_folder / "hydrographs"
    os.makedirs(export_folder, exist_ok=True)
    export = {}
    for station, rp in df_event.items():
        for rp, df in rp.items():
            if rp not in export:
                export[rp] = df["twl"].to_frame(name=station)
            else:
                export[rp] = pd.concat(
                    [export[rp], df["twl"].to_frame(name=station)], axis=1
                )
    for rp in export:
        export[rp].to_csv(export_folder / f"gtsm_normal_tide_hydrograph_rp{rp:04d}.csv")

    # save storm tide hydrograph (spring tide)
    export_spring = {}
    for station, rp in df_event_spring.items():
        for rp, df in rp.items():
            if rp not in export_spring:
                export_spring[rp] = df["twl"].to_frame(name=station)
            else:
                export_spring[rp] = pd.concat(
                    [export_spring[rp], df["twl"].to_frame(name=station)], axis=1
                )
    for rp in export_spring:
        export_spring[rp].to_csv(
            export_folder / f"gtsm_spring_tide_hydrograph_rp{rp:04d}.csv"
        )


def generate_tide_signals(
    station: int, tidepd: pd.Series, make_plot: bool = True
) -> tuple[np.ndarray, np.ndarray]:
    """Generate tidal signals for a given station.

    Args:
        station: The station ID.
        tidepd: The tidal data for the station.
        make_plot: Whether to create a plot of the tidal signals.

    Returns:
        A tuple containing the average and spring tidal signals.
    """
    tidepd = tidepd[datetime(1980, 1, 1) : datetime(2017, 12, 31, 23, 50)]
    tidepd = tidepd.dropna()  # drop NaN values
    # tidepd = tidepd.set_index("time")
    tide_array = tidepd.values  # open dataframe as numpy array

    if abs(np.quantile(tide_array, 0.99)) < abs(
        np.quantile(tide_array, 0.01)
    ):  # check whether minima or maxima have higher absolute values
        index = [
            tide_array[0:150].argmin()
        ]  # if minima are higher in absolute terms, find first minima index
        for c in itertools.count():
            search_around_index = (
                index[-1] + 6 * 24 + 5
            )  # search for minima of the next tidal cycle (24 hours 50 min. later)
            try:
                minima_index = tidepd[
                    search_around_index - 24 : search_around_index + 25
                ].values.argmin()  # search for minima between +20 hours & 50 min. and +28 hours & 50 min.
                index.append(
                    search_around_index + minima_index - 24
                )  # save index of minima
            except:
                break

        steps = []
        tidal_cycles = []
        for i in range(len(index) - 1):  # loop over index minima
            tidal_cycle = tide_array[
                index[i] - 75 : index[i + 1] - 35
            ]  # select part of time series (-75 is around previous low tide)
            if (
                len(tidal_cycle) > 150
            ):  # length of the selected tidal cycle should be at least 150 steps, otherwise the cycle is not a complete one and cannot be used to extract the average tidal cycle
                tidal_cycles.append(tidal_cycle)  # save tidal cycle
                steps.append(
                    len(tidal_cycle)
                )  # save number of steps (minimum of all saved steps becomes the length)

    else:
        index = [
            tide_array[0:150].argmax()
        ]  # if maxima are higher in absolute terms, find first maxima index
        for c in (
            itertools.count()
        ):  # search for minima of the next tidal cycle (24 hours 50 min. later)
            search_around_index = index[-1] + 6 * 24 + 5
            try:
                maxima_index = tidepd[
                    search_around_index - 24 : search_around_index + 25
                ].values.argmax()  # search for maxima between +20 hours & 50 min. and +28 hours & 50 min.
                index.append(
                    search_around_index + maxima_index - 24
                )  # save index of maxima
            except:
                break

        steps = []
        tidal_cycles = []
        for i in range(len(index) - 1):  # loop over index maxima
            tidal_cycle = tide_array[
                index[i] - 37 : index[i + 1]
            ]  # select part of time series (-37 is around previous low tide)
            if (
                len(tidal_cycle) > 150
            ):  # length of the selected tidal cycle should be at least 150 steps, otherwise the cycle is not a complete one and cannot be used to extract the average tidal cycle
                tidal_cycles.append(tidal_cycle)  # save tidal cycle
                steps.append(
                    len(tidal_cycle)
                )  # save number of steps (minimum of all saved steps becomes the length)

    length = np.min(steps)
    tides_equal_length = tidal_cycles[0][:length]
    for i in range(1, len(tidal_cycles)):
        tides_equal_length = np.vstack((tides_equal_length, tidal_cycles[i][:length]))

    tides_mean = np.mean(
        tides_equal_length, axis=0
    )  # compute the mean of all tidal cycles to get the average tidal cycle
    idx_max = tides_mean[:149].argmax() + 447
    average_tide_signal = np.tile(tides_mean[:149], 7)[idx_max - 447 : idx_max + 447]

    # plot average tide signal
    if make_plot:
        tides_min = np.min(tides_equal_length, axis=0)
        tides_max = np.max(tides_equal_length, axis=0)
        plt.fill_between(
            np.arange(298) / 6,
            np.concatenate((tides_min[:149], tides_min[:149])),
            np.concatenate((tides_max[:149], tides_max[:149])),
            fc="grey",
            alpha=0.3,
            label="tidal cycles",
        )
        plt.plot(
            np.arange(298) / 6,
            np.concatenate((tides_mean[:149], tides_mean[:149])),
            linewidth=2,
            color="black",
            linestyle="solid",
            label="average tide signal",
        )  # plot average tidal cycle as a dashed black line

    # Compute spring tide signal
    tide_array = tidepd.values
    tide_maxima_index, tide_maxima_values = ss.find_peaks(
        tide_array, distance=1728, height=-15
    )
    tide_maxima_values = tide_maxima_values["peak_heights"]
    tide_time_array = tidepd.index.values
    maxima_tide_time = tide_time_array[tide_maxima_index.tolist()]

    spring_tides = []
    for p in range(1, len(maxima_tide_time) - 1):
        spring_tide = tide_array[
            tide_maxima_index[p] - 447 : tide_maxima_index[p] + 447
        ]
        spring_tides.append(spring_tide)

    spring_tide_signal = np.mean(spring_tides, axis=0)

    # plot spring tide signal
    if make_plot:
        spring_tides_min = np.min(spring_tides, axis=0)
        spring_tides_max = np.max(spring_tides, axis=0)
        plt.fill_between(
            np.arange(225) / 6,
            spring_tides_min[335:560],
            spring_tides_max[335:560],
            fc="red",
            alpha=0.3,
            label="spring tidal cycles",
        )
        plt.plot(
            np.arange(225) / 6,
            spring_tide_signal[335:560],
            linewidth=2,
            color="red",
            linestyle="solid",
            label="spring tide signal",
        )
        plt.title(f"tidal cycles {station}")
        plt.ylabel("water level (m)")
        plt.xlabel("time (hours)")
        plt.xticks([0, 6, 12, 18, 24, 30, 36], ["0", "6", "12", "18", "24", "30", "36"])
        plt.legend(loc="upper left", fontsize=9)
        ymin, ymax = plt.ylim()
        plt.ylim(top=((ymax - ymin) * 0.25 + ymax))
        plt.xlim(0, 36)
        plt.grid()
        plt.savefig(
            "plot/gtsm/tide_signal_station_%05d.png" % (int(station)),
            format="png",
            bbox_inches="tight",
            dpi=300,
        )
        plt.close("all")
    return average_tide_signal, spring_tide_signal


def generate_surge_hydrograph(
    station: int, surgepd: pd.Series, percentile: float, make_plot: bool
) -> tuple[np.ndarray, np.ndarray]:
    """Generate storm surge hydrographs for a given GEB model.

    Args:
        station: The station ID.
        surgepd: The surge data for the station.
        percentile: The percentile to use for peak selection.
        make_plot: Whether to create plots of the hydrographs.

    Returns:
        A tuple containing the average and peak surge hydrographs arrays.
    """
    # read surge data
    surgepd = surgepd[datetime(1980, 1, 1) : datetime(2017, 12, 31, 23, 50)]
    surgepd = surgepd.dropna()  # drop NaN values

    # select surge maxima's
    distance = 432
    surge_maxima_index, surge_maxima_values = ss.find_peaks(
        surgepd.values, distance=distance, height=-10
    )
    surge_peaks = pd.DataFrame(
        data={"waterlevel": surge_maxima_values["peak_heights"]},
        index=surgepd.index.values[surge_maxima_index.tolist()],
    ).sort_values(by="waterlevel")
    surge_peaks_POT = surge_peaks[
        surge_peaks.waterlevel >= surge_peaks.quantile(percentile).waterlevel
    ]  # 0.6 #1.5 #0.8

    # generate storm surge hydrograph
    hours = 36
    df_before_peak = pd.DataFrame(index=np.around(np.arange(0, 1.0001, 0.005), 3))
    df_after_peak = pd.DataFrame(index=np.around(np.arange(0, 1.0001, 0.005), 3))
    for k in range(len(surge_peaks_POT)):
        timeseries_before_peak = surgepd.loc[
            surge_peaks_POT.iloc[k].name
            - timedelta(hours=hours) : surge_peaks_POT.iloc[k].name
        ]
        timeseries_after_peak = surgepd.loc[
            surge_peaks_POT.iloc[k].name : surge_peaks_POT.iloc[k].name
            + timedelta(hours=hours)
        ]

        normalized_before_peak = timeseries_before_peak / timeseries_before_peak.max()
        normalized_after_peak = timeseries_after_peak / timeseries_after_peak.max()

        # maybe select that part from where the values never exceed zero anymore? Or just use the whole timeslice, independent from value?
        if normalized_after_peak.min() < 0:
            select_stop = np.argwhere(normalized_after_peak.values < 0)[0][0]
        else:
            select_stop = hours * 6

        if normalized_before_peak.min() < 0:
            select_start = np.argwhere(normalized_before_peak.values < 0)[-1][0]
        else:
            select_start = 1

        # this part is only for plotting
        normalized_before_25 = normalized_before_peak[select_start:]
        normalized_after_25 = normalized_after_peak[:select_stop]
        normalized_after_25_plot = normalized_after_peak[: select_stop + 1]
        yy = np.concatenate(
            (
                normalized_before_25.values[:-1],
                normalized_after_25_plot.values,
            )
        )
        xx = (
            np.arange(-len(normalized_before_25) + 1, len(normalized_after_25_plot)) / 6
        )
        if k == 0:
            plt.plot(
                xx, yy, linewidth=0.5, color="grey", alpha=0.5, label="storm surges"
            )
        else:
            plt.plot(xx, yy, linewidth=0.5, color="grey", alpha=0.5)

        for l in df_before_peak.index.values:
            df_before_peak.loc[l, "event" + str(k)] = np.nansum(
                normalized_before_25.values > l
            )

        for l in df_after_peak.index.values:
            df_after_peak.loc[l, "event" + str(k)] = np.nansum(
                normalized_after_25.values > l
            )

    df_before_peak["mean"] = df_before_peak.mean(axis=1)
    df_before_peak["75th"] = df_before_peak.drop("mean", axis=1).quantile(
        q=0.75, axis=1
    )
    df_before_peak["25th"] = df_before_peak.drop("mean", axis=1).quantile(
        q=0.25, axis=1
    )
    df_after_peak["mean"] = df_after_peak.mean(axis=1)
    df_after_peak["75th"] = df_after_peak.drop("mean", axis=1).quantile(q=0.75, axis=1)
    df_after_peak["25th"] = df_after_peak.drop("mean", axis=1).quantile(q=0.25, axis=1)

    surge_hydrograph_height = np.concatenate(
        (
            np.zeros(247),
            np.hstack(
                (df_before_peak.index.values, np.flipud(df_after_peak.index.values)[1:])  # ty:ignore[no-matching-overload]
            ),  # ty:ignore[no-matching-overload]
            np.zeros(246),
        )
    )
    surge_hydrograph_duration_mean = np.concatenate(
        (
            np.full(247, np.nan),
            np.hstack(
                (
                    df_before_peak["mean"].values,
                    np.flipud(df_after_peak["mean"].values)[1:],
                )
            ),
            np.full(246, np.nan),
        )
    )

    if make_plot:
        plt.plot(
            -df_before_peak["mean"].values * (1 / 6),
            df_before_peak.index.values,
            label="surge hydrograph",
            color="green",
            linewidth=3,
            linestyle="--",
        )
        plt.plot(
            df_after_peak["mean"].values * (1 / 6),
            df_after_peak.index.values,  # ty:ignore[invalid-argument-type]
            color="green",
            linewidth=3,
            linestyle="--",
        )
        plt.fill_betweenx(
            df_before_peak.index.values,  # ty:ignore[invalid-argument-type]
            -df_before_peak["25th"].values * (1 / 6),
            -df_before_peak["75th"].values * (1 / 6),
            fc="green",
            alpha=0.3,
            label="P 25th-75th",
        )
        plt.fill_betweenx(
            df_after_peak.index.values,  # ty:ignore[invalid-argument-type]
            df_after_peak["25th"].values * (1 / 6),
            df_after_peak["75th"].values * (1 / 6),
            fc="green",
            alpha=0.3,
        )

        plt.legend(loc="upper left", fontsize=9)
        plt.xlabel("time relative to peak (hours)")
        plt.ylabel("normalized surge level")
        plt.title(f"surge hydrograph {station}")
        plt.ylim(0, 1.25)
        plt.xlim(-35, 35)
        plt.xticks(
            [-36, -24, -12, 0, 12, 24, 36], ["-36", "-24", "-12", "0", "12", "24", "36"]
        )
        plt.yticks(
            [0, 0.2, 0.4, 0.6, 0.8, 1, 1.2],
            ["0.0", "0.2", "0.4", "0.6", "0.8", "1.0", "1.2"],
        )
        plt.grid()
        plt.savefig(
            "plot/gtsm/surge_hydrograph_station_%05d_percentile_%02d.png"
            % (int(station), percentile * 100),
            format="png",
            bbox_inches="tight",
            dpi=300,
        )
        plt.close("all")

    return surge_hydrograph_height, surge_hydrograph_duration_mean


def generate_storm_tide_hydrograph(
    rps: pd.DataFrame,
    station: int,
    tidepd: pd.Series,
    surgepd: pd.Series,
    average_tide_signal: np.ndarray,
    spring_tide_signal: np.ndarray,
    surge_hydrograph_duration_mean: np.ndarray,
    surge_hydrograph_height: np.ndarray,
    percentile: float,
    rp: int,
    offset: int,
    make_plot: bool,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """This function generates storm tide hydrographs for a given station.

    Args:
        rps: A DataFrame containing return period statistics.
        station: The station ID.
        tidepd: A Series containing tide predictions.
        surgepd: A Series containing surge predictions.
        average_tide_signal: An array containing the average tide signal.
        spring_tide_signal: An array containing the spring tide signal.
        surge_hydrograph_duration_mean: An array containing the mean duration of the surge hydrograph.
        surge_hydrograph_height: An array containing the height of the surge hydrograph.
        percentile: The percentile to use for the hydrograph.
        rp: The return period.
        offset: The offset for the hydrograph.
        make_plot: Whether to create a plot.

    Returns:
        A tuple containing the DataFrames with the tide and surge hydrographs.

    """
    surgepd = surgepd[datetime(1980, 1, 1) : datetime(2017, 12, 31, 23, 50)]
    tidepd = tidepd[datetime(1980, 1, 1) : datetime(2017, 12, 31, 23, 50)]

    rl = rps[str(rp)].loc[int(station)]
    xvalues = np.around(np.arange(-36, 36.1, 1 / 6), 3)

    surge_height = rl - np.max(average_tide_signal)
    surge_rise = np.flip(
        np.interp(
            np.arange(0, np.nanmax(np.ceil(surge_hydrograph_duration_mean[:448])), 1),
            np.flip(surge_hydrograph_duration_mean[:448]),
            np.flip(surge_hydrograph_height[:448] * surge_height),
        )
    )
    surge_fall = np.interp(
        np.arange(
            0, np.nanmax(np.ceil(np.flipud(surge_hydrograph_duration_mean[447:]))), 1
        ),
        surge_hydrograph_duration_mean[447:],
        surge_hydrograph_height[447:] * surge_height,
    )
    surge_rise_full = np.hstack((np.zeros(448 - len(surge_rise)), surge_rise))
    surge_fall_full = np.hstack((surge_fall, np.zeros(447 - len(surge_fall))))
    surge = np.hstack((surge_rise_full, surge_fall_full[1:]))
    df_event = pd.DataFrame(
        data={
            "tide": average_tide_signal,
            "surge": surge,
            "twl": average_tide_signal + surge,
        },
        index=pd.date_range(start="1/1/2000", periods=len(surge), freq="10min"),
    )
    surge_height_spring = rl - np.max(spring_tide_signal)
    surge_rise_spring = np.flip(
        np.interp(
            np.arange(0, np.nanmax(np.ceil(surge_hydrograph_duration_mean[:448])), 1),
            np.flip(surge_hydrograph_duration_mean[:448]),
            np.flip(surge_hydrograph_height[:448] * surge_height_spring),
        )
    )
    surge_fall_spring = np.interp(
        np.arange(
            0, np.nanmax(np.ceil(np.flipud(surge_hydrograph_duration_mean[447:]))), 1
        ),
        surge_hydrograph_duration_mean[447:],
        surge_hydrograph_height[447:] * surge_height_spring,
    )
    surge_rise_full_spring = np.hstack(
        (np.zeros(448 - len(surge_rise_spring)), surge_rise_spring)
    )
    surge_fall_full_spring = np.hstack(
        (surge_fall_spring, np.zeros(447 - len(surge_fall_spring)))
    )
    surge_spring = np.hstack((surge_rise_full_spring, surge_fall_full_spring[1:]))
    df_event_spring = pd.DataFrame(
        data={
            "tide": spring_tide_signal,
            "surge": surge_spring,
            "twl": spring_tide_signal + surge_spring,
        },
        index=pd.date_range(start="1/1/2000", periods=len(surge_spring), freq="10min"),
    )

    # plot
    if make_plot:
        df_event_plot = df_event[149 : 149 * 5]
        plt.plot(
            xvalues,
            df_event_plot.tide.values[82:515],
            label="average tide signal",
            color="black",
            linestyle="-",
            zorder=9,
        )
        plt.plot(
            xvalues,
            df_event_plot.surge.values[82:515],
            label="scaled surge hydrograph",
            color="green",
            linestyle="-",
            zorder=10,
        )
        plt.plot(
            xvalues,
            df_event_plot.surge.values[82:515] / surge_height,
            label="average surge hydrograph",
            color="green",
            linestyle="--",
            zorder=10,
        )
        plt.plot(
            xvalues,
            df_event_plot.twl.values[82:515],
            label="storm tide hydrograph",
            color="tab:blue",
            zorder=8,
        )
        plt.legend(fontsize=9, loc="upper left")
        plt.xlabel("time relative to peak (hours)")
        plt.ylabel("water level (m)")
        plt.title(f"storm tide hydrograph {station}")
        plt.xticks(
            [-36, -24, -12, 0, 12, 24, 36], ["-36", "-24", "-12", "0", "12", "24", "36"]
        )
        plt.title(f"storm tide RP{rp} hydrograph {station}")
        plt.grid()
        ymin, ymax = plt.ylim()
        plt.ylim(top=((ymax - ymin) * 0.20 + ymax))
        plt.xlim(-36, 36)
        plt.savefig(
            "plot/gtsm/rp%4d_storm_tide_hydrograph_station_%05d_percentile_%02d.png"
            % (rp, int(station), percentile * 100),
            format="png",
            bbox_inches="tight",
            dpi=300,
        )
        plt.close("all")

        df_event_plot_spring = df_event_spring[149 : 149 * 5]
        plt.plot(
            xvalues,
            df_event_plot_spring.tide.values[82:515],
            label="spring tide signal",
            color="black",
            linestyle="-",
            zorder=9,
        )
        plt.plot(
            xvalues,
            df_event_plot_spring.surge.values[82:515],
            label="scaled surge hydrograph",
            color="green",
            linestyle="-",
            zorder=10,
        )
        plt.plot(
            xvalues,
            df_event_plot_spring.surge.values[82:515] / surge_height_spring,
            label="average surge hydrograph",
            color="green",
            linestyle="--",
            zorder=10,
        )
        plt.plot(
            xvalues,
            df_event_plot_spring.twl.values[82:515],
            label="storm tide hydrograph",
            color="tab:blue",
            zorder=8,
        )
        plt.legend(fontsize=9, loc="upper left")
        plt.xlabel("time relative to peak (hours)")
        plt.ylabel("water level (m)")
        plt.title(f"storm tide hydrograph {station}")
        plt.xticks(
            [-36, -24, -12, 0, 12, 24, 36], ["-36", "-24", "-12", "0", "12", "24", "36"]
        )
        plt.title(f"storm tide RP{rp} hydrograph {station}")
        plt.grid()
        ymin, ymax = plt.ylim()
        plt.ylim(top=((ymax - ymin) * 0.20 + ymax))
        plt.xlim(-36, 36)
        plt.savefig(
            "plot/gtsm/rp%4d_storm_tide_spring_tide_hydrograph_station_%05d_percentile_%02d.png"
            % (int(rp), int(station), percentile * 100),
            format="png",
            bbox_inches="tight",
            dpi=300,
        )
        plt.close("all")

    return (
        df_event,
        df_event_spring,
    )



"""Data adapter for CWATM water demand data."""

from typing import Any

import rioxarray  # noqa: F401
import xarray as xr

from .base import Adapter
from .workflows.google_drive import download_from_google_drive

FILES: dict[str, dict[str, str]] = {
    "industry": {
        "historical": "1QrDQ59qb4pa6PxnvePhCoKnlvvG2DhZO",
        "ssp1": "16t14PSPWJ7L1lJNh7NBePYmgR-RmZLX3",
        "ssp2": "1goboGVpWZh7u2qv9M8jsSnDHQ8wkuZPN",
        "ssp3": "1p-a__Wjp7sa5lIOCIcY0VZMU1BL4HWK2",
        "ssp5": "13c_Tw7wA7LPzcMXi6mvs6w-vW8W7MH2j",
    },
    "livestock": {
        "historical": "1z0fQgdF7BUGbmm07-mUAxgIt8OrRN9gC",
        "ssp2": "1L6DZ_GVZnJCmBqnRaMkfeZikD-PKo5RS",
    },
}


class CWATMWaterDemand(Adapter):
    """The CWATM Water Demand adapter for downloading water demand data."""

    def __init__(self, variable: str, *args: Any, **kwargs: Any) -> None:
        """Initialize the CWATMWaterDemand adapter.

        Args:
            variable: The variable to download, either 'industry' or 'livestock'.
            *args: Positional arguments to pass to the Adapter constructor.
            **kwargs: Keyword arguments to pass to the Adapter constructor.
        """
        self.variable = variable
        super().__init__(*args, **kwargs)

    def fetch(self, url: str | None = None) -> Adapter:
        """Download the CWATM water demand file from Google Drive.

        Args:
            url: The URL to download data from (unused, IDs are looked up in FILES).

        Returns:
            The adapter instance.

        Raises:
            ValueError: If the scenario cannot be determined or the ID is missing.
        """
        if not self.is_ready:
            # Determine scenario from filename
            # Filenames are expected to contain the scenario (e.g. historical_liv_...)
            filename = self.path.name
            scenario = next(
                (s for s in FILES[self.variable] if f"{s}_" in filename), None
            )

            if not scenario:
                raise ValueError(
                    f"Could not determine scenario from filename: {filename}. Expected one of {list(FILES[self.variable].keys())}"
                )

            file_id = FILES[self.variable][scenario]
            download_from_google_drive(file_id=file_id, file_path=self.path)

        return self

    def read(self, **kwargs: Any) -> xr.Dataset:
        """Read the CWATM water demand data.

        Args:
            **kwargs: Additional keyword arguments forwarded to xarray.

        Returns:
            The dataset.
        """
        ds = xr.open_dataset(self.path, decode_times=False, **kwargs).rename(
            {"lat": "y", "lon": "x"}
        )
        ds = ds.rio.write_crs("EPSG:4326")
        return ds


class CWATMLivestockWaterDemand(CWATMWaterDemand):
    """The CWATM Livestock Water Demand adapter."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the CWATMLivestockWaterDemand adapter."""
        super().__init__("livestock", *args, **kwargs)


class CWATMIndustryWaterDemand(CWATMWaterDemand):
    """The CWATM Industry Water Demand adapter."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the CWATMIndustryWaterDemand adapter."""
        super().__init__("industry", *args, **kwargs)



"""Workflow utilities for handling geometry data."""

from pathlib import Path
from typing import Any

import geopandas as gpd
from shapely.geometry import MultiPolygon, Polygon


def read_parquet_with_geom(
    path: Path, geom: Polygon | MultiPolygon | None = None, **kwargs: Any
) -> gpd.GeoDataFrame:
    """Read a parquet file with optional spatial filtering.

    Args:
        path: Path to the parquet file.
        geom: Optional shapely Polygon to filter the geometries.
        **kwargs: Additional keyword arguments to pass to gpd.read_parquet.

    Returns:
        A GeoDataFrame with the data from the parquet file, optionally filtered by the geometry.
    """
    if geom is not None:
        assert "bbox" not in kwargs, "Cannot use both geom and bbox"
        assert isinstance(geom, (Polygon, MultiPolygon)), (
            "geom must be a Polygon or MultiPolygon"
        )
        gdf: gpd.GeoDataFrame = gpd.read_parquet(path, **kwargs, bbox=geom.bounds)
        gdf: gpd.GeoDataFrame = gdf[gdf.intersects(geom)]
        return gdf
    else:
        return gpd.read_parquet(path, **kwargs)



"""Utilities to download DeltaDTM tiles for a given bounding box.

This module provides a downloader that downloads and extracts the needed
GeoTIFF tiles from the remote ZIP packages hosted by DeltaDTM.
Tiles are 10x10-degree and downloaded individually.

Notes:
    - DeltaDTM distributes 10x10-degree tiles as individual TIFF files zipped into ZIP archives per continent.
      Tile filenames follow the pattern "DeltaDTM_v1_1_N00W000.tif" where the coordinates
      indicate the bounding box of the tile. See the DeltaDTM documentation for details.
    - Coverage spans the global land areas.

"""

from __future__ import annotations

import os
import tempfile
import zipfile
from pathlib import Path
from typing import Any

import geopandas as gpd
import numpy as np
import rioxarray as rxr
import xarray as xr
from rioxarray import merge

from geb.workflows.io import fetch_and_save
from geb.workflows.raster import convert_nodata

from .base import Adapter

available_continents = {
    "Africa.zip": "https://data.4tu.nl/file/1da2e70f-6c4d-4b03-86bd-b53e789cc629/22ffa027-184b-4f67-9979-c182f3dfb1ab",
    "Antarctica.zip": "https://data.4tu.nl/file/1da2e70f-6c4d-4b03-86bd-b53e789cc629/ca957a40-34fa-41eb-b101-e45d1ccbd890",
    "Asia.zip": "https://data.4tu.nl/file/1da2e70f-6c4d-4b03-86bd-b53e789cc629/672eba4c-1334-44c6-8119-8879ded25912",
    "Europe.zip": "https://data.4tu.nl/file/1da2e70f-6c4d-4b03-86bd-b53e789cc629/cb0b8ee3-b018-4828-a74e-2fb05020b1b6",
    "North_America.zip": "https://data.4tu.nl/file/1da2e70f-6c4d-4b03-86bd-b53e789cc629/037664c6-1494-4889-9689-a56570728320",
    "Oceania.zip": "https://data.4tu.nl/file/1da2e70f-6c4d-4b03-86bd-b53e789cc629/de972de1-26bd-4303-afdf-21a90a232cff",
    "Seven_seas_(open_ocean).zip": "https://data.4tu.nl/file/1da2e70f-6c4d-4b03-86bd-b53e789cc629/fe986ba6-3db9-40e2-8a49-0fcdb341244a",
    "South_America.zip": "https://data.4tu.nl/file/1da2e70f-6c4d-4b03-86bd-b53e789cc629/db980f00-63cd-4a07-a4df-55ab06510594",
}


class DeltaDTM(Adapter):
    """Downloader for DeltaDTM tiles.

    Downloads and extracts the needed GeoTIFF tiles from the remote ZIP packages
    hosted by DeltaDTM for a given bounding box.

    Attributes:
        cache_dir (Path): Directory to cache downloaded tiles.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the adapter for DeltaDTM.

        Args:
            *args: Additional positional arguments passed to the base Adapter class.
            **kwargs: Additional keyword arguments passed to the base Adapter class.
        """
        super().__init__(*args, **kwargs)

    def get_tiles_in_model_bounds(
        self, xmin: float, xmax: float, ymin: float, ymax: float
    ) -> tuple[list[str], list[str]]:
        """Get the DeltaDTM tiles that intersect with the model bounds. This function uses the DeltaDTM tiles geopackage.

        Args:
            xmin: Minimum x-coordinate (longitude) of the model bounds.
            xmax: Maximum x-coordinate (longitude) of the model bounds.
            ymin: Minimum y-coordinate (latitude) of the model bounds.
            ymax: Maximum y-coordinate (latitude) of the model bounds.
        Returns:
            A tuple containing:
                - A list of tile filenames that intersect with the model bounds.
                - A list of continent ZIP filenames to download.
        Raises:
            RuntimeError: If the DeltaDTM tiles geopackage cannot be downloaded.
        """
        # download the DeltaDTM tiles geopackage
        url_delta_dtm_tiles = "https://data.4tu.nl/file/1da2e70f-6c4d-4b03-86bd-b53e789cc629/60a69899-2e67-4f9f-8761-3b57094acd12"
        os.makedirs(self.root, exist_ok=True)
        with tempfile.TemporaryDirectory() as temp_dir_str:
            temp_dir: Path = Path(temp_dir_str)
            filepath = temp_dir / "delta_dtm_tiles.gpkg"
            success = fetch_and_save(
                url=url_delta_dtm_tiles,
                file_path=filepath,
            )
            if not success:
                raise RuntimeError("Failed to download DeltaDTM tiles geopackage.")
            # load the geopackage
            gdf_tiles = gpd.read_file(filepath)
        # get the tiles that intersect with the model bounds
        tiles_in_bounds = gdf_tiles.cx[xmin:xmax, ymin:ymax]
        tile_names = tiles_in_bounds["tile"].tolist()

        # continents(s) to download tiles for
        continents_to_download = tiles_in_bounds["zipfile"].unique().tolist()

        return tile_names, continents_to_download

    def download_deltadtm(self, continents_to_download: list[str]) -> None:
        """Download DeltaDTM tiles for the specified continents.

        Args:
            continents_to_download: List of continent ZIP filenames to download.
        Raises:
            RuntimeError: If downloading any of the continent ZIP files fails.
        """
        for continent in continents_to_download:
            url = available_continents[continent]
            zip_path = self._construct_filepath(continent[:-4])
            success = fetch_and_save(
                url=url,
                file_path=zip_path,
            )
            if not success:
                raise RuntimeError(
                    f"Failed to download DeltaDTM continent ZIP: {continent}"
                )

    def unpack_and_merge_tiles(
        self, continents_to_download: list[str], tile_names: list[str]
    ) -> xr.DataArray:
        """Unpack and merge DeltaDTM tiles into a single dataarray.

        Args:
            continents_to_download: List of continent ZIP filenames to extract from.
            tile_names: List of tile filenames to unpack and merge.

        Returns:
            Merged dataarray of the specified tiles.
        """
        with tempfile.TemporaryDirectory() as temp_dir_str:
            temp_dir: Path = Path(temp_dir_str)
            extracted_paths = self._unpack_tiles(
                continents_to_download, tile_names, temp_dir
            )
            da = self._merge_tiles(extracted_paths)

        da = convert_nodata(da, np.nan)
        return da

    def _merge_tiles(self, tile_paths: list[Path]) -> xr.DataArray:
        """Merge extracted DeltaDTM tiles into a single xarray DataArray.

        Args:
            tile_paths: List of paths to the extracted tile files.

        Returns:
            Merged dataset of the specified tiles.
        """
        das: list[xr.DataArray] = [rxr.open_rasterio(path) for path in tile_paths]  # ty: ignore[invalid-assignment]
        das = [da.sel(band=1) for da in das]
        da: xr.DataArray = merge.merge_arrays(das)
        return da

    def _unpack_tiles(
        self, continents_to_download: list[str], tile_names: list[str], temp_dir: Path
    ) -> list[Path]:
        """Unpack and merge DeltaDTM tiles into a single xarray Dataset.

        Args:
            continents_to_download: List of continent ZIP filenames to extract from.
            tile_names: List of tile filenames to unpack and merge.
            temp_dir: Temporary directory to extract tiles into.
        Returns:
            Merged dataset of the specified tiles.
        """
        extracted_paths: list[Path] = []
        for continent in continents_to_download:
            zip_path = self._construct_filepath(continent[:-4])
            with zipfile.ZipFile(zip_path, "r") as zip_ref:
                for tile_name in tile_names:
                    if tile_name in zip_ref.namelist():
                        zip_ref.extract(tile_name, path=temp_dir)
                        extracted_paths.append(temp_dir / tile_name)
        return extracted_paths

    def _construct_filepath(self, continent: str) -> Path:
        """Construct the file path for a given continent ZIP file.

        Args:
            continent: The continent ZIP filename.
        Returns:
            The constructed file path for the continent ZIP file.

        """
        return Path(str(self.path).format(continent))

    def fetch(
        self, xmin: float, xmax: float, ymin: float, ymax: float, url: str | None = None
    ) -> DeltaDTM:
        """Fetch DeltaDTM tiles for the specified bounding box.

        Args:
            xmin: Minimum x-coordinate (longitude) of the bounding box.
            xmax: Maximum x-coordinate (longitude) of the bounding box.
            ymin: Minimum y-coordinate (latitude) of the bounding box.
            ymax: Maximum y-coordinate (latitude) of the bounding box.
            url: URL to download DeltaDTM data from. Defaults to None.
        Returns:
            The DeltaDTM adapter instance with the downloaded data.
        """
        self.tile_names, self.continents_to_download = self.get_tiles_in_model_bounds(
            xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax
        )
        self.download_deltadtm(self.continents_to_download)
        return self

    def read(self) -> xr.DataArray:
        """Read and unpack the downloaded DeltaDTM data.

        Returns:
            The downloaded DeltaDTM data.
        """
        da = self.unpack_and_merge_tiles(self.continents_to_download, self.tile_names)
        return da



"""Lakes and Reservoirs Module."""

# --------------------------------------------------------------------------------
# This file contains code that has been adapted from an original source available
# in a public repository under the GNU General Public License. The original code
# has been modified to fit the specific needs of this project.
#
# Original source repository: https://github.com/iiasa/CWatM
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------------------

from __future__ import annotations

from typing import TYPE_CHECKING

import geopandas as gpd
import numpy as np
import pandas as pd

from geb.geb_types import Array, ArrayBool, ArrayFloat32, ArrayFloat64, ArrayInt32
from geb.module import Module
from geb.store import Bucket
from geb.workflows import balance_check
from geb.workflows.io import read_geom, read_grid

if TYPE_CHECKING:
    from geb.model import GEBModel, Hydrology

OFF: int = 0
LAKE: int = 1
RESERVOIR: int = 2
LAKE_CONTROL: int = 3  # currently modelled as normal lake


GRAVITY: np.float32 = np.float32(9.81)
SHAPE: str = "parabola"
# http://rcswww.urz.tu-dresden.de/~daigner/pdf/ueberf.pdf

if SHAPE == "rectangular":
    overflow_coefficient_mu: np.float32 = np.float32(0.577)

    def estimate_lake_outflow(
        lake_factor: ArrayFloat32,
        height_above_outflow: ArrayFloat32,
    ) -> ArrayFloat32:
        """Estimates the outflow from a lake given its height above the outflow for a rectangular shape.

        References:
            http://rcswww.urz.tu-dresden.de/~daigner/pdf/ueberf.pdf

        Args:
            lake_factor: A lake-specific constant factor used in the lake outflow equations.
            height_above_outflow: Height of the lake above elevation of the outflow [m]

        Returns:
            Outflow in m3/s
        """
        return lake_factor * height_above_outflow**1.5

    def outflow_to_height_above_outflow(
        lake_factor: ArrayFloat32, outflow: ArrayFloat32
    ) -> ArrayFloat32:
        """Inverse function of estimate_lake_outflow for a rectangular shape.

        Args:
            lake_factor: A lake-specific constant factor used in the lake outflow equations.
            outflow: Outflow in m3/s

        Returns:
            Height of the lake above elevation of the outflow [m]
        """
        return (outflow / lake_factor) ** np.float32((2 / 3))

elif SHAPE == "parabola":
    overflow_coefficient_mu: np.float32 = np.float32(0.612)

    def estimate_lake_outflow(
        lake_factor: ArrayFloat32,
        height_above_outflow: ArrayFloat32,
    ) -> ArrayFloat32:
        """Estimates the outflow from a lake given its height above the outflow for a parabolic shape.

        References:
            http://rcswww.urz.tu-dresden.de/~daigner/pdf/ueberf.pdf

        Args:
            lake_factor: A lake-specific constant factor used in the lake outflow equations.
            height_above_outflow: Height of the lake above the outflow [m]

        Returns:
            Outflow in m3/s
        """
        return lake_factor * height_above_outflow**2

    def outflow_to_height_above_outflow(
        lake_factor: ArrayFloat32, outflow: ArrayFloat32
    ) -> ArrayFloat32:
        """Inverse function of estimate_lake_outflow for a parabolic shape.

        References:
            http://rcswww.urz.tu-dresden.de/~daigner/pdf/ueberf.pdf

        Args:
            lake_factor: A lake-specific constant factor used in the lake outflow equations.
            outflow: Outflow in m3/s

        Returns:
            Height of the lake above the outflow [m]
        """
        return np.sqrt(outflow / lake_factor)

else:
    raise ValueError("Invalid shape")


def get_lake_height_from_bottom(
    lake_storage: ArrayFloat64, lake_area: ArrayFloat32
) -> ArrayFloat32:
    """Calculate the height of a lake above the bottom given its storage and area.

    Assumes a box-shaped lake. Could be extended in the future to account for different lake shapes.

    Args:
        lake_storage: Storage of the lake in m3
        lake_area: Area of the lake in m2

    Returns:
        Height of the lake above the bottom in m
    """
    height_from_bottom = (lake_storage / lake_area).astype(np.float32)
    return height_from_bottom


def get_lake_storage_from_height_above_bottom(
    lake_height: ArrayFloat32, lake_area: ArrayFloat32
) -> ArrayFloat64:
    """Calculate the storage of a lake given its height above the bottom and area.

    Args:
        lake_height: Height of the lake above the bottom in m
        lake_area: Area of the lake in m2

    Returns:
        Storage of the lake in m3
    """
    return (lake_height * lake_area).astype(np.float64)


def get_lake_height_above_outflow(
    lake_storage: ArrayFloat64,
    lake_area: ArrayFloat32,
    outflow_height: ArrayFloat32,
) -> ArrayFloat32:
    """Calculate the height of a lake above the outflow given its storage, area, and outflow height.

    Assumes a box-shaped lake. Could be extended in the future to account for different lake shapes.
    First calculates the height above the bottom, then subtracts the outflow height.

    Args:
        lake_storage: Storage of the lake in m3
        lake_area: Area of the lake in m2
        outflow_height: Height of the outflow in m above the bottom of the lake in m (assuming a rectangular lake)

    Returns:
        Height of the lake above the outflow in m
    """
    height_above_outflow = (
        get_lake_height_from_bottom(lake_storage, lake_area) - outflow_height
    )
    height_above_outflow[height_above_outflow < 0] = 0
    return height_above_outflow


def get_river_width(
    average_discharge: ArrayFloat32,
) -> ArrayFloat32:
    """Estimate river width at lake outflow from average discharge using an empirical relationship.

    TODO: Check if this can be improved using river width data from the hydrological model.

    Args:
        average_discharge: Average discharge in m3/s

    Returns:
        Estimated river width in m
    """
    return np.float32(7.1) * np.power(average_discharge, np.float32(0.539))


def get_lake_factor(
    river_width: ArrayFloat32,
    overflow_coefficient_mu: np.float32,
    lake_a_factor: np.float32,
) -> ArrayFloat32:
    """A lake-constant factor that is used in the equations for lake outflow.

    Pre-calculated to save computation time.

    References:
        http://rcswww.urz.tu-dresden.de/~daigner/pdf/ueberf.pdf

    Args:
        river_width: Width of the river in m
        overflow_coefficient_mu: Overflow coefficient, depends on the shape of the lake outflow
        lake_a_factor: Calibration factor for the lake outflow

    Returns:
        A lake-specific constant factor used in the lake outflow equations.
    """
    return (
        lake_a_factor
        * overflow_coefficient_mu
        * np.float32((2 / 3))
        * river_width
        * np.float32((2 * GRAVITY) ** 0.5)
    )


def estimate_outflow_height(
    lake_capacity: ArrayFloat64,
    lake_factor: ArrayFloat32,
    lake_area: ArrayFloat32,
    avg_outflow: ArrayFloat32,
) -> ArrayFloat32:
    """Estimate the outflow height of a lake given its capacity, lake factor, area, and average outflow.

    Args:
        lake_capacity: Capacity of the lake in m3
        lake_factor: Factor for the Modified Puls approach to calculate retention of the lake
        lake_area: Area of the lake in m2
        avg_outflow: Average outflow in m3/s

    Returns:
        outflow_height: Height of the outflow in m above the bottom of the lake in m (assuming a rectangular lake)

    """
    height_above_outflow = outflow_to_height_above_outflow(lake_factor, avg_outflow)
    lake_height_when_full = get_lake_height_from_bottom(lake_capacity, lake_area)
    outflow_height = lake_height_when_full - height_above_outflow
    outflow_height[outflow_height < 0] = 0
    return outflow_height


def get_lake_outflow(
    dt: float,
    storage: ArrayFloat64,
    lake_factor: ArrayFloat32,
    lake_area: ArrayFloat32,
    outflow_height: ArrayFloat32,
) -> tuple[ArrayFloat32, ArrayFloat32]:
    """Calculate outflow and storage for a lake using the Modified Puls method.

    Args:
        dt: Time step in seconds
        storage: Current storage in m3
        lake_factor: Factor for the Modified Puls approach to calculate retention of the lake
        lake_area: Area of the lake in m2
        outflow_height: Height of the outflow in m above the bottom of the lake in m (assuming a rectangular lake)

    Returns:
        outflow_m3: Outflow in m3.
        height_above_outflow: Height of the lake above the outflow in m.
    """
    height_above_outflow = get_lake_height_above_outflow(
        lake_storage=storage, lake_area=lake_area, outflow_height=outflow_height
    )
    storage_above_outflow = height_above_outflow * lake_area
    storage_above_outflow = np.minimum(storage_above_outflow, storage)

    outflow_m3_s = estimate_lake_outflow(lake_factor, height_above_outflow)
    outflow_m3 = outflow_m3_s * dt

    outflow_m3 = np.minimum(outflow_m3, storage_above_outflow)

    assert (outflow_m3 <= storage).all()
    return outflow_m3, height_above_outflow


class WaterBodyVariables(Bucket):
    """Variables for the Lakes and Reservoirs module."""

    storage: ArrayFloat64
    lake_area: ArrayFloat32
    lake_factor: ArrayFloat32
    waterbody_mapping: ArrayInt32
    waterbody_ids_original: ArrayInt32
    waterbody_data: pd.DataFrame
    capacity: ArrayFloat64
    waterbody_type: ArrayInt32
    outflow_height: ArrayFloat32


class WaterBodies(Module):
    """Implements all lakes and reservoir operations in the hydrological model.

    For reservoir it gets the outflow from the reservoir operator agents.

    Args:
        model: The GEB model instance.
        hydrology: The hydrology submodel instance.
    """

    var: WaterBodyVariables

    def __init__(self, model: GEBModel, hydrology: Hydrology) -> None:
        """Initializes the Lakes and Reservoirs module.

        Args:
            model: The GEB model instance.
            hydrology: The hydrology submodel instance.
        """
        super().__init__(model)
        self.hydrology = hydrology

        self.HRU = hydrology.HRU
        self.grid = hydrology.grid

        if self.model.in_spinup:
            self.spinup()

    @property
    def name(self) -> str:
        """Name of the module.

        Returns:
            The name of the module.
        """
        return "hydrology.waterbodies"

    def spinup(self) -> None:
        """Spinup part to initialize lakes and reservoirs.

        This function is called during the spinup phase of the model to set up lakes and reservoirs.

        Steps:
        1. Load the water bodies map and create a mapping of water body IDs.
        2. Set discharge to NaN for all cells that are not part of a water body.
        3. Identify outflow points for each water body.
        4. Load water body data from a Parquet file and filter it based on the water bodies present in the grid.
        5. Extract relevant attributes such as water body type, area, capacity, and average discharge.
        6. Calculate river width and lake factor for each water body.
        7. Initialize storage for lakes and reservoirs, setting initial values based on capacity.
        8. Estimate outflow height for each water body.

        """
        # load lakes/reservoirs map with a single ID for each lake/reservoir
        waterBodyID_unmapped: np.ndarray = self.grid.load(
            self.model.files["grid"]["waterbodies/waterbody_id"]
        )
        self.grid.var.waterBodyID, self.var.waterbody_mapping = (
            self.map_waterbodies_IDs(waterBodyID_unmapped)
        )

        # set discharge to NaN for all cells that are not part of a water body
        self.grid.var.discharge_in_rivers_m3_s_substep[
            self.grid.var.waterBodyID != -1
        ] = np.nan

        self.grid.var.waterbody_outflow_points = self.get_outflows(
            self.grid.var.waterBodyID
        )

        # we compress the waterbody_outflow_points, which we can later use to decompress
        waterbody_ids = np.unique(
            self.grid.var.waterbody_outflow_points[
                self.grid.var.waterbody_outflow_points != -1
            ]
        )

        self.var.waterbody_data = self.load_waterbody_data(
            self.var.waterbody_mapping, waterBodyID_unmapped
        )
        # sort the water bodies in the same order as the compressed water body IDs (waterbody_ids)
        self.var.waterbody_data = self.var.waterbody_data.sort_index()

        assert np.array_equal(self.var.waterbody_data.index, waterbody_ids)

        self.var.waterbody_type = self.var.waterbody_data["waterbody_type"].values
        self.var.waterbody_ids_original = self.var.waterbody_data[
            "original_waterbody_id"
        ].values
        # change water body type to LAKE if it is a control lake, thus currently modelled as normal lake
        self.var.waterbody_type[self.var.waterbody_type == LAKE_CONTROL] = LAKE

        # print("setting all water body types to LAKE")
        # self.var.waterbody_type.fill(LAKE)

        assert (np.isin(self.var.waterbody_type, [LAKE, RESERVOIR])).all()

        self.var.lake_area = self.var.waterbody_data["average_area"].values
        self.var.capacity = self.var.waterbody_data["volume_total"].values

        # lake discharge at outlet to calculate alpha: parameter of channel width, gravity and weir coefficient
        # Lake parameter A (suggested  value equal to outflow width in [m])
        average_discharge = np.maximum(
            self.var.waterbody_data["average_discharge"].values,
            0.1,
        )

        # channel width in [m]
        river_width = get_river_width(average_discharge)

        self.var.lake_factor = get_lake_factor(
            river_width,
            overflow_coefficient_mu,
            self.model.config["parameters"]["lake_outflow_multiplier"],
        )

        self.var.storage = np.full_like(self.var.capacity, np.nan, dtype=np.float64)

        # initialize storage to 50% of the capacity. This is arbitrary, but
        # ok since we use a spinup period
        self.reservoir_storage = self.reservoir_capacity * 0.5
        self.var.outflow_height = estimate_outflow_height(
            self.var.capacity,
            self.var.lake_factor,
            self.var.lake_area,
            average_discharge,
        )

        # initialize lake storage to exactly the level of the outflow height
        self.lake_storage = get_lake_storage_from_height_above_bottom(
            self.var.outflow_height[self.is_lake], self.var.lake_area[self.is_lake]
        )
        assert (
            get_lake_height_above_outflow(
                self.lake_storage,
                self.var.lake_area[self.is_lake],
                self.var.outflow_height[self.is_lake],
            )
            < 1e-5
        ).all()

    def map_waterbodies_IDs(
        self, waterBodyID_unmapped: ArrayInt32
    ) -> tuple[ArrayInt32, ArrayInt32]:
        """Maps the water body IDs to a continuous range of IDs starting from 0.

        If there are no water bodies, it returns an array of -1.

        Args:
            waterBodyID_unmapped: The original water body IDs from the grid.

        Returns:
            A tuple containing:
                - waterBodyID_mapped: The mapped water body IDs.
                - waterbody_mapping: The mapping from original IDs to mapped IDs.
        """
        unique_waterbodies = np.unique(waterBodyID_unmapped)
        unique_waterbodies = unique_waterbodies[unique_waterbodies != -1]
        if unique_waterbodies.size == 0:
            return np.full_like(waterBodyID_unmapped, -1), np.full(
                1, -1, dtype=np.int32
            )
        else:
            waterbody_mapping = np.full(
                unique_waterbodies.max() + 2, -1, dtype=np.int32
            )  # make sure that the last entry is also -1, so that -1 maps to -1
            waterbody_mapping[unique_waterbodies] = np.arange(
                0, unique_waterbodies.size, dtype=np.int32
            )
            return waterbody_mapping[waterBodyID_unmapped], waterbody_mapping

    def load_waterbody_data(
        self,
        waterbody_mapping: ArrayInt32,
        waterbody_original_ids: ArrayInt32,
    ) -> gpd.GeoDataFrame:
        """Loads water body data from a Parquet file and sets the index to the mapped water body IDs.

        All water bodies that are not in the original IDs (and thus not in the study area) are dropped.

        Args:
            waterbody_mapping: The mapping from original water body IDs to mapped IDs.
            waterbody_original_ids: The original water body IDs from the grid.

        Returns:
            A GeoDataFrame containing the water body data with the index set to the mapped water body IDs.
        """
        waterbody_data = read_geom(
            self.model.files["geom"]["waterbodies/waterbody_data"],
        )
        # drop all data that is not in the original ids
        waterbody_original_ids_compressed = np.unique(waterbody_original_ids)
        waterbody_original_ids_compressed = waterbody_original_ids_compressed[
            waterbody_original_ids_compressed != -1
        ]
        waterbody_data = waterbody_data[
            waterbody_data["waterbody_id"].isin(waterbody_original_ids_compressed)
        ]
        # map the waterbody ids to the new ids, save old ids
        waterbody_data["original_waterbody_id"] = waterbody_data["waterbody_id"]
        waterbody_data["waterbody_id"] = waterbody_mapping[
            waterbody_data["waterbody_id"]
        ]

        waterbody_data = waterbody_data.set_index("waterbody_id")
        return waterbody_data

    def get_outflows(self, waterBodyID: ArrayInt32) -> ArrayInt32:
        """Identifies the outflow points for each water body.

        Finds the cell with the highest upstream area in each water body as the outflow point.
        If there are multiple cells with the same upstream area, the one with the lowest elevation is chosen.

        Args:
            waterBodyID: The mapped water body IDs from the grid.

        Returns:
            An array containing the outflow point for each water body.
        """
        # calculate biggest outlet = biggest accumulation of ldd network
        upstream_area_n_cells = self.hydrology.routing.grid.var.upstream_area_n_cells
        upstream_area_within_waterbodies = np.zeros_like(
            upstream_area_n_cells,
            shape=waterBodyID.max() + 2,
        )
        upstream_area_within_waterbodies[-1] = -1
        np.maximum.at(
            upstream_area_within_waterbodies,
            waterBodyID[waterBodyID != -1],
            upstream_area_n_cells[waterBodyID != -1],
        )
        upstream_area_within_waterbodies = np.take(
            upstream_area_within_waterbodies, waterBodyID
        )

        # in some cases the cell with the highest number of upstream cells
        # has mulitple occurences in the same lake, this seems to happen
        # especially for very small lakes with a small drainage area.
        # In such cases, we take the outflow cell with the lowest elevation.
        outflow_elevation = read_grid(
            self.model.files["grid"]["landsurface/elevation_min_m"]
        )
        outflow_elevation = self.grid.compress(outflow_elevation)

        waterbody_outflow_points = np.where(
            upstream_area_n_cells == upstream_area_within_waterbodies,
            waterBodyID,
            -1,
        )

        number_of_outflow_points_per_waterbody = np.unique(
            waterbody_outflow_points, return_counts=True
        )
        duplicate_outflow_points = number_of_outflow_points_per_waterbody[0][
            number_of_outflow_points_per_waterbody[1] > 1
        ]
        duplicate_outflow_points = duplicate_outflow_points[
            duplicate_outflow_points != -1
        ]

        if duplicate_outflow_points.size > 0:
            # in some cases the cell with the highest number of upstream cells
            # has mulitple occurences in the same lake, this seems to happen
            # especially for very small lakes with a small drainage area.
            # In such cases, we take the outflow cell with the lowest elevation.
            outflow_elevation = self.grid.compress(
                read_grid(self.model.files["grid"]["landsurface/elevation_min_m"])
            )

            for duplicate_outflow_point in duplicate_outflow_points:
                duplicate_outflow_points_indices = np.where(
                    waterbody_outflow_points == duplicate_outflow_point
                )[0]
                minimum_elevation_outflows_idx = np.argmin(
                    outflow_elevation[duplicate_outflow_points_indices]
                )
                non_minimum_elevation_outflows_indices = (
                    duplicate_outflow_points_indices[
                        (
                            np.arange(duplicate_outflow_points_indices.size)
                            != minimum_elevation_outflows_idx
                        )
                    ]
                )
                waterbody_outflow_points[non_minimum_elevation_outflows_indices] = -1

        if __debug__:
            # make sure that each water body has an outflow
            assert np.array_equal(
                np.unique(waterbody_outflow_points), np.unique(waterBodyID)
            )
            # make sure that each outflow point is only used once
            unique_outflow_points = np.unique(
                waterbody_outflow_points[waterbody_outflow_points != -1],
                return_counts=True,
            )[1]
            if unique_outflow_points.size > 0:
                assert unique_outflow_points.max() == 1

        return waterbody_outflow_points

    def routing_lakes(self, routing_step_length_seconds: int | float) -> ArrayFloat32:
        """Lake routine to calculate lake outflow.

        Args:
            routing_step_length_seconds: length of the routing step in seconds.

        Returns:
            lake_outflow_m3: Outflow from the lakes in m3 per routing step.

        """
        is_lake = self.is_lake
        # check if there are any lakes in the model
        if is_lake.any():
            (
                lake_outflow_m3,
                _,
            ) = get_lake_outflow(
                routing_step_length_seconds,
                self.var.storage[is_lake],
                self.var.lake_factor[is_lake],
                self.var.lake_area[is_lake],
                self.var.outflow_height[is_lake],
            )
        else:
            lake_outflow_m3 = np.zeros(0, dtype=np.float32)

        return lake_outflow_m3

    def routing_reservoirs(
        self, n_routing_substeps: int, current_substep: int
    ) -> tuple[ArrayFloat32, ArrayFloat32]:
        """Routine to update reservoir volumes and calculate reservoir outflow.

        Args:
            n_routing_substeps: Number of routing substeps per time step
            current_substep: Current substep in the routing process

        Returns:
            main_channel_release_m3: Outflow from the main channel in m3 per routing step.
            command_area_release_m3: Outflow from the command area in m3 per routing step
        """
        main_channel_release_m3, command_area_release_m3 = (
            self.model.agents.reservoir_operators.release(
                daily_substeps=n_routing_substeps,
                current_substep=current_substep,
            )
        )

        assert (main_channel_release_m3 <= self.reservoir_storage).all()
        assert (self.reservoir_storage >= 0).all()

        return main_channel_release_m3, command_area_release_m3

    def substep(
        self,
        current_substep: int,
        n_routing_substeps: int,
        routing_step_length_seconds: int,
    ) -> tuple[ArrayFloat32, ArrayFloat32]:
        """Routes lakes and reservoirs for a single routing substep.

        Importantly, the outflow is not removed from the storage here. This is
        done in the hydrology routing module.

        For lakes, it calculates the outflow based on the current storage and lake parameters.
        This is returned as outflow to the drainage network.

        For reservoirs, it gets the outflow and the command area release from the reservoir operators.
        The outflow is returned as outflow to the drainage network, while the command area release
        is returned separately.

        The command area release is not routed through the drainage network, but is instead
        added directly to the irrigation demand in the irrigation module.

        Args:
            current_substep: Current substep in the routing process
            n_routing_substeps: Number of routing substeps per time step
            routing_step_length_seconds: length of the routing step in seconds.

        Returns:
            tuple containing:
                - outflow_to_drainage_network_m3: Outflow to the drainage network in m3 per routing step.
                - command_area_release_m3: Outflow from the command area in m3 per routing step.
        """
        if __debug__:
            prestorage = self.var.storage.copy()

        outflow_to_drainage_network_m3 = np.zeros_like(
            self.var.storage, dtype=np.float32
        )
        command_area_release_m3 = np.zeros_like(self.var.storage, dtype=np.float32)

        outflow_to_drainage_network_m3[self.is_lake] = self.routing_lakes(
            routing_step_length_seconds
        )

        (
            outflow_to_drainage_network_m3[self.is_reservoir],
            command_area_release_m3[self.is_reservoir],
        ) = self.routing_reservoirs(n_routing_substeps, current_substep)

        assert (
            outflow_to_drainage_network_m3 <= self.var.storage.astype(np.float32)
        ).all(), (
            f"Outflow exceeds storage: {outflow_to_drainage_network_m3.max()} > {self.var.storage.max()}"
        )

        if __debug__:
            balance_check(
                name="command_area_release",
                how="cellwise",
                influxes=[],
                outfluxes=[],
                prestorages=[prestorage[self.is_reservoir]],
                poststorages=[self.var.storage[self.is_reservoir]],
                tolerance=1,  # 1 m3
            )

        return outflow_to_drainage_network_m3, command_area_release_m3

    @property
    def is_reservoir(self) -> ArrayBool:
        """Returns a boolean array indicating which water bodies are reservoirs.

        Returns:
            A boolean array where True indicates that the corresponding water body is a reservoir.
        """
        return self.var.waterbody_type == RESERVOIR

    @property
    def is_lake(self) -> ArrayBool:
        """Returns a boolean array indicating which water bodies are lakes.

        Returns:
            A boolean array where True indicates that the corresponding water body is a lake.
        """
        return self.var.waterbody_type == LAKE

    @property
    def reservoir_storage(self) -> ArrayFloat64:
        """Gets the storage of each reservoir in the model.

        Returns:
            The storage of each reservoir in the model. [m3]
        """
        return self.var.storage[self.is_reservoir]

    @reservoir_storage.setter
    def reservoir_storage(self, value: ArrayFloat64) -> None:
        """Sets the storage of each reservoir in the model."""
        self.var.storage[self.is_reservoir] = value

    @property
    def reservoir_capacity(self) -> ArrayFloat64:
        """Gets the capacity of each reservoir in the model.

        Returns:
            The capacity of each reservoir in the model. [m3]
        """
        return self.var.capacity[self.is_reservoir]

    @reservoir_capacity.setter
    def reservoir_capacity(self, value: ArrayFloat64) -> None:
        """Sets the capacity of each reservoir in the model.

        Args:
            value: The capacity of each reservoir in the model. [m3]
        """
        self.var.capacity[self.is_reservoir] = value

    @property
    def lake_storage(self) -> ArrayFloat64:
        """Gets the storage of each lake in the model.

        Returns:
            The storage of each lake in the model. [m3]
        """
        return self.var.storage[self.is_lake]

    @lake_storage.setter
    def lake_storage(self, value: ArrayFloat64) -> None:
        """Sets the storage of each lake in the model."""
        self.var.storage[self.is_lake] = value

    @property
    def lake_capacity(self) -> ArrayFloat64:
        """Gets the capacity of each lake in the model.

        Returns:
            The capacity of each lake in the model. [m3]
        """
        return self.var.capacity[self.is_lake]

    @lake_capacity.setter
    def lake_capacity(self, value: ArrayFloat64) -> None:
        """Sets the capacity of each lake in the model."""
        self.var.capacity[self.is_lake] = value

    @property
    def reservoir_fill_percentage(self) -> ArrayFloat32:
        """Returns the fill percentage of each reservoir in the model.

        Returns:
            The fill percentage of each reservoir in the model.
        """
        return (self.reservoir_storage / self.reservoir_capacity * 100).astype(
            np.float32
        )

    @property
    def n(self) -> int:
        """Returns the number of lakes and reservoirs in the model.

        Returns:
            The number of lakes and reservoirs in the model.
        """
        return self.var.capacity.size

    def decompress(self, array: Array) -> Array:
        """Placeholder decompression function.

        Other modules expect a decompress function to be present. However, since
        the lakes and reservoirs module doesn't need to decompress any data,
        this function simply returns the input array unchanged.

        Args:
            array: The input array to be "decompressed".

        Returns:
            The same array that was input, unchanged.
        """
        return array

    def step(self) -> None:
        """Dynamic part set lakes and reservoirs for each year."""
        # if first timestep, or beginning of new year
        if self.model.current_timestep == 1 or (
            self.model.current_time.month == 1 and self.model.current_time.day == 1
        ):
            if self.hydrology.dynamic_waterbodies:
                raise NotImplementedError("dynamic_waterbodies not implemented yet")

        # print(self.reservoir_fill_percentage.astype(int))
        self.report(locals())



"""Module for livestock farmers agent-based simulation."""

from __future__ import annotations

import calendar
import datetime
from typing import TYPE_CHECKING, Literal

import numpy as np
import numpy.typing as npt
import xarray as xr

from geb.geb_types import ArrayFloat32
from geb.hydrology.HRUs import load_water_demand_xr
from geb.store import Bucket

from ..hydrology.landcovers import GRASSLAND_LIKE
from .general import AgentBaseClass, downscale_volume

if TYPE_CHECKING:
    from geb.agents import Agents
    from geb.model import GEBModel


class LiveStockFarmersVariables(Bucket):
    """Variables for the LivestockFarmers agent."""

    current_water_demand: ArrayFloat32
    current_efficiency: np.float32
    last_water_demand_update: datetime.datetime


class LiveStockFarmers(AgentBaseClass):
    """This class is used to simulate the government.

    Args:
        model: The GEB model.
        agents: The class that includes all agent types (allowing easier communication between agents).
    """

    var: LiveStockFarmersVariables

    def __init__(
        self, model: GEBModel, agents: Agents, reduncancy: float | None
    ) -> None:
        """Initialize the LivestockFarmers agent module.

        Note that currently, this module is not actually agent-based but rather
        uses aggregated pre-defined water demand data.

        Args:
            model: The GEB model.
            agents: The class that includes all agent types (allowing easier communication between agents).
            reduncancy: Number of extra agents to use. Not used here.
        """
        super().__init__(model)

        if self.model.simulate_hydrology:
            self.HRU = model.hydrology.HRU
            self.grid = model.hydrology.grid

        self.agents = agents
        self.config = (
            self.model.config["agent_settings"]["town_managers"]
            if "town_managers" in self.model.config["agent_settings"]
            else {}
        )

        self.livestock_water_consumption_ds: xr.Dataset = load_water_demand_xr(
            self.model.files["other"]["water_demand/livestock_water_consumption"]
        )

        if self.model.in_spinup:
            self.spinup()

    @property
    def name(self) -> str:
        """The name of the module.

        Used to save data to disk.

        Returns:
            The name of the module.
        """
        return "agents.livestock_farmers"

    def spinup(self) -> None:
        """Set initial water demand values during spinup."""
        water_demand, efficiency = self.update_water_demand()
        self.var.current_water_demand = water_demand
        self.var.current_efficiency = efficiency

    def update_water_demand(self) -> tuple[ArrayFloat32, np.float32]:
        """Update the water demand for livestock farmers at the HRU level.

        Returns:
            A tuple containing:
            - The updated water demand as a numpy array (in m3/day).
            - The current efficiency as a float.
        """
        days_in_year: Literal[366, 365] = (
            366 if calendar.isleap(self.model.current_time.year) else 365
        )

        # grassland/non-irrigated land that is not owned by a crop farmer
        land_use_type = self.HRU.var.land_use_type
        downscale_mask = (land_use_type != GRASSLAND_LIKE) | (
            self.HRU.var.land_owners != -1
        )

        # transform from mio m3 per year to m3/day
        water_consumption = (
            self.livestock_water_consumption_ds.sel(
                time=self.model.current_time,
                method="ffill",
                tolerance="366D",  # ty:ignore[invalid-argument-type]
            ).livestock_water_consumption
            * 1_000_000
            / days_in_year
        )
        water_consumption = (
            water_consumption.rio.write_crs(4326).rio.reproject(
                4326,
                shape=self.model.hydrology.grid.shape,
                transform=self.model.hydrology.grid.transform,
            )
            / (
                water_consumption.rio.transform().a
                / self.model.hydrology.grid.transform.a
            )
            ** 2
        )
        water_consumption: npt.NDArray[np.float32] = (
            downscale_volume(
                water_consumption.rio.transform().to_gdal(),
                self.model.hydrology.grid.gt,
                water_consumption.values,
                self.model.hydrology.grid.mask,
                self.model.hydrology.mapping_grid_to_HRU_uncompressed,
                downscale_mask,
                self.HRU.var.land_use_ratio,
            )
            / self.HRU.var.cell_area
        )  # convert to m/day

        efficiency: np.float32 = np.float32(1.0)
        water_demand = water_consumption / efficiency
        self.var.last_water_demand_update = self.model.current_time
        return water_demand, efficiency

    def water_demand(self) -> tuple[ArrayFloat32, np.float32]:
        """Get the current water demand for livestock farmers at the HRU level.

        Updates the water demand only if data for this timestep is available.
        Otherwise, assumes the last known water demand.

        Returns:
            A tuple containing:
            - The current water demand as a numpy array (in m3/day).
            - The current efficiency as a float.
        """
        if (
            np.datetime64(self.model.current_time, "ns")
            in self.livestock_water_consumption_ds.time
        ):
            water_demand, efficiency = self.update_water_demand()
            self.var.current_water_demand = water_demand
            self.var.current_efficiency = efficiency

        assert (
            self.model.current_time - self.var.last_water_demand_update
        ).days < 366, (
            "Water demand has not been updated for over a year. "
            "Please check the livestock water demand datasets."
        )
        return self.var.current_water_demand, self.var.current_efficiency

    def step(self) -> None:
        """This function is run each timestep."""
        self.report(locals())



"""Adapter for OECD datasets."""

from __future__ import annotations

from geb.workflows.io import fetch_and_save

from .base import Adapter


class OECD(Adapter):
    """Adapter for OECD datasets."""

    def fetch(self, url: str) -> OECD:
        """Fetch the dataset from the given URL if not already present.

        Args:
            url: The URL to fetch the dataset from.

        Returns:
            The OECD adapter instance.
        """
        if not self.is_ready:
            fetch_and_save(url=url, file_path=self.path)
        return self



"""Data adapter for obtaining ERA5 data from the Destination Earth."""

from __future__ import annotations

import base64
import os
import time
from datetime import datetime, timedelta
from typing import Any

import aiohttp
import numpy as np
import xarray as xr

from geb.workflows.raster import convert_nodata, interpolate_na_along_dim

from .base import Adapter

N_CONNECTION_ATTEMPTS = 3
RETRY_DELAY_SECONDS = 5


class DestinationEarth(Adapter):
    """Data adapter for obtaining ERA5 data from the Destination Earth."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the ERA5 data adapter."""
        super().__init__(*args, **kwargs)

    def get_authentication_header(self) -> dict[str, str]:
        """Generate the authentication header for Destination Earth access.

        Returns:
            A dictionary containing the Authorization header for HTTP requests.

        Raises:
            ValueError: If the DESTINATION_EARTH_KEY environment variable is not set.
        """
        DESTINATION_EARTH_KEY: str | None = os.getenv(key="DESTINATION_EARTH_KEY")
        if DESTINATION_EARTH_KEY is None:
            print("ERROR: DESTINATION_EARTH_KEY environment variable is not set.")
            print(
                "Please set your Personal Access Token in your .env file or export it in your shell."
            )
            raise ValueError("DESTINATION_EARTH_KEY environment variable is not set.")

        auth_string: str = f"edh:{DESTINATION_EARTH_KEY}"
        # Base64 encode the "username:password" string (edh:<PAT>)
        encoded_auth: str = base64.b64encode(auth_string.encode("utf-8")).decode(
            "utf-8"
        )

        auth_headers: dict[str, str] = {"Authorization": f"Basic {encoded_auth}"}
        return auth_headers

    def fetch(self, url: str) -> DestinationEarth:
        """Set the URL for the Destination Earth data source.

        Args:
            url: The URL of the Destination Earth data source.

        Returns:
            The current instance of the DestinationEarth adapter.
        """
        self.url = url
        return self

    def connect_API(
        self,
        variable: str,
        start_date: datetime,
        end_date: datetime,
        bounds: tuple[float, float, float, float],
    ) -> xr.DataArray:
        """Download ERA5 data for a specific variable and time period.

        Args:
            variable: Short name of the variable to download (e.g., "t2m"). Codes can be found here: https://codes.ecmwf.int/grib/param-db/
            start_date: start date of the time period to download
            end_date: end date of the time period to download
            bounds:  bounding box in the format (min_lon, min_lat, max_lon, max_lat)

        Returns:
            Downloaded ERA5 data as an xarray DataArray.

        Raises:
            ConnectionError: If unable to connect to the Destination Earth API after multiple attempts.
        """
        for attempt in range(N_CONNECTION_ATTEMPTS):
            try:
                da: xr.DataArray = xr.open_dataset(
                    self.url,
                    storage_options={"headers": self.get_authentication_header()},
                    chunks={},
                    engine="zarr",
                )[variable].rename(
                    {"valid_time": "time", "latitude": "y", "longitude": "x"}
                )
                break
            except aiohttp.ClientResponseError:
                print(
                    f"Error connecting to Destination Earth API. This could be due to erroneous credentials or a temporary server issue. Retrying ({attempt}/{N_CONNECTION_ATTEMPTS})..."
                )
                time.sleep(RETRY_DELAY_SECONDS)
        else:
            raise ConnectionError(
                "Failed to connect to Destination Earth API after 3 attempts."
            )

        da: xr.DataArray = da.drop_vars(
            ["number", "surface", "depthBelowLandLayer"], errors="ignore"
        )

        buffer: float = 0.5
        buffered_bounds: tuple[float, float, float, float] = (
            bounds[0] - buffer,
            bounds[1] - buffer,
            bounds[2] + buffer,
            bounds[3] + buffer,
        )

        # Check if region crosses the meridian (longitude=0)
        # use a slightly larger slice. The resolution is 0.1 degrees, so buffer degrees is a bit more than that (to be sure)
        if buffered_bounds[0] < 0 and buffered_bounds[2] > 0:
            # Need to handle the split across the meridian
            # Get western hemisphere part (longitude < 0)
            west_da: xr.DataArray = da.sel(
                time=slice(start_date, end_date),
                y=slice(buffered_bounds[3], buffered_bounds[1]),
                x=slice(((buffered_bounds[0]) + 360) % 360, 360),
            )
            # Get eastern hemisphere part (longitude > 0)
            east_da: xr.DataArray = da.sel(
                time=slice(start_date, end_date),
                y=slice(buffered_bounds[3], buffered_bounds[1]),
                x=slice(0, ((buffered_bounds[2]) + 360) % 360),
            )
            # Combine the two parts
            da: xr.DataArray = xr.concat([west_da, east_da], dim="x")
        else:
            # Regular case - doesn't cross meridian
            da: xr.DataArray = da.sel(
                time=slice(start_date, end_date),
                y=slice(buffered_bounds[3], buffered_bounds[1]),
                x=slice(
                    ((buffered_bounds[0]) + 360) % 360,
                    ((buffered_bounds[2]) + 360) % 360,
                ),
            )

        # Reorder x to be between -180 and 180 degrees
        da: xr.DataArray = da.assign_coords(x=((da.x + 180) % 360 - 180))

        assert da.x.size > 0 and da.y.size > 0, (
            "No data found for the specified bounds."
        )

        da.attrs["_FillValue"] = da.attrs["GRIB_missingValue"]
        da: xr.DataArray = convert_nodata(da, np.nan)
        return da

    def read(
        self,
        variable: str,
        start_date: datetime,
        end_date: datetime,
        bounds: tuple[float, float, float, float],
    ) -> xr.DataArray:
        """Process ERA5 data for a given variable and time period.

        Downloads the data from the Climate Data Store (CDS) if not already available,
        processes it to ensure it is in the correct format, and applies de-accumulation
        for accumulated variables and interpolation of missing values.

        Args:
            variable: short name of the variable to process (e.g., "t2m"). Codes can be found here: https://codes.ecmwf.int/grib/param-db/
            start_date: start date of the time period to process
            end_date: end date of the time period to process
            bounds:  bounding box in the format (min_lon, min_lat, max_lon, max_lat)

        Raises:
            NotImplementedError: If the step type of the data is not "accum" or "instant".

        Returns:
            xr.DataArray: Processed ERA5 data as an xarray DataArray.
        """
        da: xr.DataArray = self.connect_API(
            variable, start_date - timedelta(hours=1), end_date, bounds
        )
        # assert that time is monotonically increasing with a constant step size
        assert (
            da.time.diff("time").astype(np.int64)
            == (da.time[1] - da.time[0]).astype(np.int64)
        ).all(), "time is not monotonically increasing with a constant step size"
        if da.attrs["GRIB_stepType"] == "accum":
            da: xr.DataArray = xr.where(
                da.isel(time=slice(1, None)).time.dt.hour == 1,
                da.isel(time=slice(1, None)),
                da.diff(dim="time", n=1),
            )

        elif da.attrs["GRIB_stepType"] == "instant":
            da = da.isel(time=slice(1, None))
        else:
            raise NotImplementedError

        assert da.time.dt.hour.min().item() == 0, "time does not start at hour 0"

        # rechunk to have all data for a time step in one chunk
        da = da.chunk({"x": -1, "y": -1, "time": 24})

        da: xr.DataArray = da.rio.write_crs(4326)
        da: xr.DataArray = interpolate_na_along_dim(da)

        return da



"""Data adapter for SWORD data."""

import shutil
import sqlite3
import tempfile
import time
import zipfile
from pathlib import Path
from typing import Any

import geopandas as gpd
import pandas as pd

from geb.workflows.io import fetch_and_save

from .base import Adapter

FILES: list[str] = []


class Sword(Adapter):
    """The SWORD adapter for downloading and processing SWORD data."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the SWORD adapter.

        Args:
            *args: Positional arguments to pass to the Adapter constructor.
            **kwargs: Keyword arguments to pass to the Adapter constructor.
        """
        super().__init__(*args, **kwargs)

    def fetch(self, url: str) -> Adapter:
        """Process SWORD zip file to extract and convert to parquet.

        Args:
            url: The URL to download the SWORD zip file from.

        Returns:
            The adapter instance with the processed parquet file.
        """
        if not self.is_ready:
            download_path: Path = self.root / url.split(sep="/")[-1]
            fetch_and_save(url=url, file_path=download_path)

            uncompressed_file: Path = download_path.with_suffix(suffix="")
            with zipfile.ZipFile(file=download_path, mode="r") as zip_ref:
                zip_ref.extractall(uncompressed_file)

            download_path.unlink()  # remove zip file

            gdfs: list[gpd.GeoDataFrame] = []
            downloaded_files: list[Path] = []
            for continent in ("sa", "na", "oc", "eu", "af", "as"):
                reaches: Path = (
                    self.root
                    / "SWORD_v16_gpkg"
                    / "gpkg"
                    / f"{continent}_sword_reaches_v16.gpkg"
                )

                gdfs.append(gpd.read_file(reaches))

            SWORD = pd.concat(gdfs, ignore_index=True)
            temp_path = Path(tempfile.gettempdir()) / "sword.gpkg"
            SWORD.to_file(temp_path)

            time.sleep(5)  # wait a bit to ensure all file handles are closed

            # Connect to the temporary GPKG
            conn = sqlite3.connect(temp_path)
            cursor = conn.cursor()

            # Create an index on COMID
            cursor.execute("""
                CREATE INDEX IF NOT EXISTS idx_reach_id
                ON sword (reach_id);
            """)

            conn.commit()
            conn.close()

            # Move the temporary file to the final path
            shutil.move(temp_path, self.path)

            time.sleep(5)  # wait a bit to ensure all file handles are closed

            shutil.rmtree(
                path=uncompressed_file, ignore_errors=True
            )  # remove uncompressed folder

        return self
